/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Sept 2001      */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 
 
#include "oroborus.h" 

// number of steps for iconbars
#define ICONSTEPS_FAST 4
#define ICONSTEPS_MED 7
#define ICONSTEPS_SLOW 20

// not currently used -would be for hiding/showing of iconbars, but it's too slow...
#define SLIDESTEP 45

// Initial value for number of iconbar steps -use the medium one (really needs to read a pref file)
int iconbar_steps=ICONSTEPS_MED, icontab_width=5;
Client *focus_before_hide;

/* 
 * 
 * Handle different window gravities 
 * 
 */ 

void gravitate_client(Client *c, int grav) 
{ 
  int dx=0, dy=0; 
 
  sprintf(lastsub,"TRACE3: gravitate_client (client: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: gravitate_client (client: %#lx)\n",c); 
#endif 
 
  switch(GRAVITY) 
  { 
  case NorthWestGravity: 
    dx=LW; 
    dy=TH; 
    break; 
  case NorthGravity: 
    dx=(LW+RW)/2; 
    dy=TH; 
    break; 
  case NorthEastGravity: 
    dx=-RW; 
    dy=TH; 
    break; 
  case WestGravity: 
    dx = LW; 
    dy=(TH+BH)/2; 
    break; 
  case EastGravity: 
    dx=-RW; 
    dy=(TH+BH)/2; 
    break;  
  case SouthWestGravity: 
    dx=LW; 
    dy=-BH; 
    break; 
  case SouthGravity: 
    dx=(LW+RW)/2; 
    dy=-BH; 
    break; 
  case SouthEastGravity: 
    dx=-RW; 
    dy=-BH; 
    break; 
  case CenterGravity: 
    dx=(LW+RW)/2; 
    dy=(TH+BH)/2; 
    break; 
  } 
  c->x=c->x+(dx*grav); 
  c->y=c->y+(dy*grav); 
} 

/*
 * Remove this client from parent's sibling list
 */

void remove_sibling_client(Client *c)
{
  Client *cp, *cnxt, *cprv;
  
  sprintf(lastsub,"TRACE3: remove_sibling_client (client: %#lx)\n",c);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: remove_sibling_client (client: %#lx)\n",c);
#endif

  cp=c->parent;
  if(!cp) return;
  
  cnxt=cp->first_child;

/* This should never really happen! */
  if(cnxt==NULL) {
#ifdef SHOW_WEIRD
    fprintf(stderr,"%s warning: parent lost children!\n",PROGNAME);
#endif
    cp->child_count=0;
    return;
  }
  
/* This should never really happen! */
  if(cp->child_count!=0)
    cp->child_count--;
#ifdef SHOW_WEIRD
  else
    fprintf(stderr,"%s warning: parent has inconsistent child_count!\n",PROGNAME);
#endif

  if(cnxt==c)
    cp->first_child=c->next_sibling;
    
  cnxt=c->next_sibling;
  cprv=c->prev_sibling;
  
  if(cnxt)
    cnxt->prev_sibling=cprv;
  if(cprv)
    cprv->next_sibling=cnxt;
}

/*
 * Add this client to parent's sibling list
 */

void add_sibling_client(Client *c)
{
  Client *cp, *cnxt;
  
  sprintf(lastsub,"TRACE3: add_sibling_client (client: %#lx)\n",c);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: add_sibling_client (client: %#lx)\n",c);
#endif

  cp=c->parent;
  if(!cp) return;
  
  cnxt=cp->first_child;
  
  if(cnxt) {
    while(cnxt->next_sibling) {
      cnxt=cnxt->next_sibling;
    }
    cnxt->next_sibling=c;
    c->prev_sibling=cnxt;
    c->next_sibling=NULL;
  } else {
    cp->first_child=c;
    c->next_sibling=NULL;
    c->prev_sibling=NULL;
  }
  cp->child_count++;
}

/* 
 * 
 * Create a new frame part 
 * 
 */ 

Window create_frame_part(Window win, int x, int y, int w, int h, int mapit) 
{ 
  Window new;   
 
  sprintf(lastsub,"TRACE3: create_frame_part %#lx (%#lx)\n",win,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: create_frame_part %#lx (%#lx)\n",win,lastfocus); 
#endif 
  
  if(w<1 || h<1) return 0;
  
  new=XCreateSimpleWindow(dpy, win, x, y, w, h, 0, 0, 0);

  XSelectInput(dpy, new, ButtonPressMask|ButtonReleaseMask); 
  if( mapit ) XMapWindow(dpy, new); 
  return new; 
} 

void reorder_dockicon_winlist(Client *c)
{
  long i;
  int classlen;
  Client *ac;
  
  sprintf(lastsub,"TRACE2: reorder_dockicon_winlist %#lx (%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: reorder_dockicon_winlist %#lx (%#lx)\n",c,lastfocus); 
#endif

  if(c->osxwinfo.winref==0) return;
  if(c->clientclass==NULL) return;
  
  classlen=strlen(c->clientclass);
// Loop through the clients, looking for matches with class of latest client
  ac=client_list;
  for(i=0;i<client_count;i++) {
    if(ac->osxwinfo.winref!=0)
      if(ac->clientclass)
        if(strlen(ac->clientclass)==classlen)
          if(!strncmp(ac->clientclass,c->clientclass,classlen))
            ReorderOSXWin(&(ac->osxwinfo));
    ac=ac->next;
  }
}


/* 
 * 
 * Create a new client frame 
 * 
 */ 

void create_new_client(Window w, int reposition) 
{ 
  Client *c=NULL; 
  XWindowAttributes attr; 
  XSetWindowAttributes wattr; 
  PropMwmHints *mhints=NULL; 
  XWMHints *hints=NULL; 
  long dummy; 
  XID tmpwin;
  XWindowChanges wc; 
  int x, y, xmax, ymax, xmin, ymin, head=0, showfront=0; 
  XClassHint xch;
 
  sprintf(lastsub,"TRACE1: create_new_client %#lx (%#lx)\n",w,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: create_new_client %#lx (%#lx)\n",w,lastfocus); 
#endif 
 
  if((c=malloc(sizeof *c))==NULL)
  {
    if(on_cmd_line)
        fprintf(stderr, "%s: Unable to allocate memory for new client.\n", PROGNAME);
    exit(1);
  }
 
  /* Initialize client values */ 
  c->window=w; 
  c->ignore_unmap=0; 
  c->hide_state=BUTTON_UNPRESSED; 
  c->icon_state=BUTTON_UNPRESSED;
  c->close_state=BUTTON_UNPRESSED; 
  c->can_max=True; 
  c->has_frame=True; 
  c->has_focus=False; 
// assume it isn't an iconbar
  c->is_icon=NOT_ICONBAR;
// start with some "bad" values...
  c->name_x=-1;
  c->name_y=-1;
  c->name_w=-1;
  c->name_h=-1;
  c->icon_x=-10000;
  c->icon_y=-10000;
  c->icon_h=-1;
  c->icon_w=-1;
  c->icontxt_x=-1;
  c->icontxt_y=-1;
  c->icontxt_w=-1;
  c->icontxt_h=-1;
// default to a modeless window
  c->modal_state=MODELESS;
  c->sloppy_family=FOCUS_CLICK;
  c->auto_properties=AUTO_NONE;
// set up initial sibling/child list
  c->child_count=0;
  c->parent=NULL;
  c->first_child=NULL;
  c->next_sibling=NULL;
  c->prev_sibling=NULL;
// initialise the OS X window identifier
  c->osxwinfo.winref=0;
// Start the string info pointing to nothing
  c->clientcmd=NULL;
  c->clientclass=NULL;

  XGrabServer(dpy);
  /* Set event mask for client */ 
  XSelectInput(dpy, c->window, PropertyChangeMask|FocusChangeMask); 
 
#ifdef I18N_SUPPORT 
  if(use_fontset)
    /* call wrapper function for I18N */ 
    XFetchNameWrapper(dpy, c->window, &c->name); 
  else
    XFetchName(dpy, c->window, &c->name);
#else 
  XFetchName(dpy, c->window, &c->name); 
#endif /* I18N */ 

  c->size=XAllocSizeHints(); 
  XGetWMNormalHints(dpy, w, c->size, &dummy); 
 
  /* Get size and position of client */ 
  XGetWindowAttributes(dpy, c->window, &attr); 
   
  /* Handle clients with border set > 0 and adjust position accordingly */ 
  if(attr.border_width>0) 
  { 
    switch(GRAVITY) 
    { 
    case SouthEastGravity: 
      attr.x=attr.x+attr.border_width*2; 
      attr.y=attr.y+attr.border_width*2; 
      break; 
    case SouthWestGravity: 
    case SouthGravity: 
      attr.y=attr.y+attr.border_width*2; 
      break; 
    case NorthEastGravity: 
    case EastGravity: 
      attr.x=attr.x+attr.border_width*2; 
      break; 
    } 
  } 
  c->x=attr.x; 
  c->y=attr.y; 
  c->width=attr.width; 
  c->height=attr.height; 
 
#ifdef SHOW_HINTS
  printf("==== NEW WINDOW: %#lx      hints...\n",c->window);
#endif
  head=XGetClassHint(dpy,c->window,&xch);
  if(head!=0) {
#ifdef SHOW_HINTS
    printf("  res_name: %s\n",xch.res_name);
    printf("  res_class: %s\n",xch.res_class);
#endif
    c->clientcmd=strdup(xch.res_name);
    c->clientclass=strdup(xch.res_class);
    XFree(xch.res_name);
    XFree(xch.res_class);
  }
  
  hints=XGetWMHints(dpy, c->window); 
  if(hints) 
  { 
    /* Sets the initial window state if set by client */ 
    if(hints->flags&WindowGroupHint) {
      tmpwin=hints->window_group;
// assuming this what the window_group is normally used for -i.e. it points to parent window ID
      c->parent=get_client_from_window((Window) tmpwin,WINDOW_LE);
#ifdef SHOW_HINTS
      printf("Got parent client: %#lx\n",c->parent);
#endif
      if(c->parent!=NULL)
        add_sibling_client(c);
    }
    if((hints->flags & StateHint)!=0 && attr.map_state!=IsViewable) 
      set_wm_state(c->window, hints->initial_state); 
    if((hints->flags & IconPixmapHint)!=0) {
      MakeX11Window(&(c->osxwinfo),c->name,c->clientclass);
      showfront=1;
    }
#ifdef SHOW_HINTS
    printf("flags: %#lx\n",hints->flags);
    printf("input: %d\n",hints->input);
    printf("XID  : %#lx\n",hints->window_group);
#endif
    XFree(hints); 
  } else {
// set initial client state
    if(attr.map_state!=IsViewable)
      set_wm_state(c->window, NormalState);
// Need to create a new OS X window, so name appears in dock icon's window list
    MakeX11Window(&(c->osxwinfo),c->name,c->clientclass);
    //c->osxwinref=MakeWindow(c->name,c->clientclass,0);
// And bring OroborOSX/XDarwin in front so we don't have to go searching for the new window
    showfront=1;
  }
   
  /* Get MWM hints */ 
  mhints=get_mwm_hints(c->window); 
  if(mhints) 
  { 
#ifdef SHOW_HINTS
    printf("---- MWM HINTS...\n");
    printf("flags      : %#lx\n",mhints->flags);
    printf("functions  : %#lx\n",mhints->functions);
    printf("decorations: %#lx\n",mhints->decorations);
    printf("inputMode  : %#lx\n",mhints->inputMode);
    printf("status     : %#lx\n",mhints->status);
    printf("DECOR_BORDER, DECOR_TITLE: %#lx , %#lx\n",MWM_DECOR_BORDER,MWM_DECOR_TITLE); 
#endif
    if(mhints->flags==0) {
      showfront=1;
      if(c->osxwinfo.winref==0)
        MakeX11Window(&(c->osxwinfo),c->name,c->clientclass);
        //c->osxwinref=MakeWindow(c->name,c->clientclass,0);
    }
    if((mhints->flags&MWM_HINTS_DECORATIONS)!=0 && (mhints->decorations&MWM_DECOR_ALL)==0) 
    { 
      /* Window has no border if *neither* MWM_DECOR_BORDER *nor* MWM_DECOR_TITLE are set */ 
      c->has_frame=((mhints->decorations&(MWM_DECOR_BORDER|MWM_DECOR_TITLE))!=0); 

      /* Set if client can be maximized or not */ 
      c->can_max=((mhints->decorations&MWM_DECOR_MAXIMIZE)!=0); 
    }
    // make a note of modal windows... 
    if((mhints->flags&MWM_HINTS_INPUT_MODE)!=0) {
      if((mhints->inputMode&MWM_INPUT_SYSTEM_MODAL)!=0) {
        c->osxwinfo.winref=-1;
        c->modal_state=SYS_MODAL;
        if(c->parent)
          (c->parent)->modal_state|=HAS_SYS_MODAL_CHILD;
      }
      if((mhints->inputMode&MWM_INPUT_PRIMARY_APPLICATION_MODAL)!=0) {
        c->osxwinfo.winref=-1;
        c->modal_state=APP_MODAL;
        if(c->parent)
          (c->parent)->modal_state|=HAS_APP_MODAL_CHILD;
      }
    } 
    if(c->osxwinfo.winref==0) {
      // create OS X window, and bring OroborOSX/XDarwin in front
      MakeX11Window(&(c->osxwinfo),c->name,c->clientclass);
      //c->osxwinref=MakeWindow(c->name,c->clientclass,0);
      showfront=1;
    }
    XFree(mhints); 
  } else if(c->osxwinfo.winref==0) {
    // create OS X window, and bring OroborOSX/XDarwin in front
    MakeX11Window(&(c->osxwinfo),c->name,c->clientclass);
    //c->osxwinref=MakeWindow(c->name,c->clientclass,0);
    showfront=1;
  }

// This is the case when new window should not be listed in the dock's window list
  if(c->osxwinfo.winref==-1)
    c->osxwinfo.winref=0;

#ifdef SHAPE_SUPPORT 
  /* Select shape events */ 
  if(shape) 
    XShapeSelectInput(dpy, c->window, ShapeNotifyMask); 
#endif /* SHAPE */ 
 
  /* Create frame window */ 
  wattr.override_redirect=True;
  // Added the LeaveWindowMask to this, so we know when we've left the window
  wattr.event_mask=SubstructureNotifyMask|SubstructureRedirectMask|ButtonPressMask|ButtonReleaseMask|EnterWindowMask|LeaveWindowMask; 
  c->frame=XCreateWindow(dpy, root, FX, FY, FW, FH, 0, DefaultDepth(dpy, screen), CopyFromParent, DefaultVisual(dpy, screen), CWOverrideRedirect|CWEventMask, &wattr);
  // Set the cursor we want to see when inside the window -would be nice if this was more like the Mac's
  // normal cursor...
  XDefineCursor(dpy, c->frame, normal_curs); 

  /* This ignores the unmap event when reparenting */ 
  if(attr.map_state==IsViewable) 
    c->ignore_unmap++; 
 
  /* Reparent client window */ 
  XReparentWindow(dpy, c->window, c->frame, LW, TH); 
    
  /* Create frame parts if has border and/or title */ 
    // Top frame parts are needed for iconbars, even if no frame
    c->title=create_frame_part(c->frame, TX, TY, TW<1?1:TW, TH, c->has_frame); 
    c->top_left=create_frame_part(c->frame, TLX, TLY, TLW, TLH, c->has_frame); 
    c->top_right=create_frame_part(c->frame, TRX, TRY, TRW, TRH, c->has_frame);  
   
  if(c->has_frame)
  { 
    /* Side frame parts */ 
    c->left=create_frame_part(c->frame, LX, LY, LW, LH, 1); 
    c->right=create_frame_part(c->frame, RX, RY, RW, RH, 1); 
     
    /* Bottom frame parts */ 
    c->bottom=create_frame_part(c->frame, BX, BY, BW<1?1:BW, BH, 1); 
    c->bottom_left=create_frame_part(c->frame, BLX, BLY, BLW, BLH, 1); 
    c->resize=create_frame_part(c->frame, RSX, RSY, RSW, RSH, 1); 
   
    /* Top frame parts 
    c->title=create_frame_part(c->frame, TX, TY, TW<1?1:TW, TH); 
    c->top_left=create_frame_part(c->frame, TLX, TLY, TLW, TLH); 
    c->top_right=create_frame_part(c->frame, TRX, TRY, TRW, TRH);   */
   
    /* Buttons */ 
    c->hide=create_frame_part(c->frame, HX, HY, HW, HH, 1); 
    c->icon=create_frame_part(c->frame, IX, IY, IW, IH, 1);   
    c->close=create_frame_part(c->frame, CX, CY, CW, CH, 1);
  } 
 
  /* Move the client to position set by hints  -changed this to work more like twm */ 
  if((c->size->flags & USPosition)!=0 || ((c->size->flags & PPosition)!=0 && (c->x!=0 || c->y!=0)) || !reposition) 
  {
    printf("Found hints for geometry...   %d x %d  @ %d,%d\n",c->width,c->height,c->x,c->y);
#ifdef XINERAMA_SUPPORT
    if(xinerama)
    {
      /* If window positioned within 15 pixels of center of display, consider it to be centered on the current head */
      if(abs((c->x+(c->width/2))-(DisplayWidth(dpy, screen)/2))<15 && abs((c->y+(c->height/2))-(DisplayHeight(dpy, screen)/2))<15)
      {
        get_mouse_xy(&x, &y);
        head=find_head(x, y);
        c->x=screen_x(head)+(screen_width(head)/2)-(c->width/2);
        c->y=screen_y(head)+(screen_height(head)/2)-(c->height/2);
        printf("XINERAMA: Found cursor on head %d... set position to %d,%d\n",head,c->x,c->y);
      }
      head=find_head(c->x, c->y);
      printf("XINERAMA: Window top-left is on head %d\n",head);
    }
#endif /* XINERAMA */
 
    xmin=screen_x(head)+LW;
    ymin=screen_y(head)+TH;
    if( screen_y(head)==0 && gMenubarSpace!=0 )
        ymin+=22;
    xmax=(screen_x(head)+screen_width(head))-c->width-RW; 
    ymax=(screen_y(head)+screen_height(head))-c->height-BH; 
   
    printf("Got bounds (min/max): %d/%d , %d/%d\n",xmin,xmax,ymin,ymax);
    gravitate_client(c, APPLY_GRAVITY); 
    wc.x=c->x; 
    wc.y=c->y; 
 
    /* Keep window fully inside desktop */ 
    if(wc.x>xmax) wc.x=xmax; 
    if(wc.y>ymax) wc.y=ymax; 
    if(wc.x<xmin) wc.x=xmin; 
    if(wc.y<ymin) wc.y=ymin; 
  } 
 
  /* Otherwise follow window placement policy */ 
  else 
  { 
#ifdef XINERAMA_SUPPORT
    if(xinerama)
    {
      get_mouse_xy(&x, &y);
      head=find_head(x, y);
      printf("XINERAMA: Found cursor on head %d...\n",head);
    }
#endif /* XINERAMA */

    xmin=screen_x(head)+LW+config.left_margin;
    ymin=screen_y(head)+TH+config.top_margin; 
    if(screen_y(head)==0 && gMenubarSpace!=0)
        ymin+=22;
    xmax=(screen_x(head)+screen_width(head))-c->width-RW-config.right_margin; 
    ymax=(screen_y(head)+screen_height(head))-c->height-BH-config.bottom_margin; 

    printf("Got bounds (min/max): %d/%d , %d/%d\n",xmin,xmax,ymin,ymax);

    if(config.window_placement==CENTER_ROOT) 
    { 
      x=screen_x(head)+(screen_width(head)/2); 
      y=screen_y(head)+(screen_height(head)/2);
      printf("Placing in center: %d,%d\n",x,y);
    } 
    else
    {
      get_mouse_xy(&x, &y); 
      printf("Placing with cursor: %d,%d\n",x,y);
    }
    wc.x=x-(c->width/2); 
    wc.y=y-(c->height/2); 
 
    /* Keep window fully inside desktop */ 
    if(wc.x>xmax) wc.x=xmax; 
    if(wc.y>ymax) wc.y=ymax; 
    if(wc.x<xmin) wc.x=xmin; 
    if(wc.y<ymin) wc.y=ymin; 
  } 
  wc.height=c->height; 
  wc.width=c->width; 
  wc.border_width=0;
// Add one for the 'jig' to work around the blank windows problem
//  wc.x+=1;

  /* Add client to client list */ 
  if(client_list==NULL) 
  { 
    client_list=c; 
    c->prev=c; 
    c->next=c; 
  } 
  else 
  { 
    c->prev=client_list->prev; 
    c->next=client_list; 
    client_list->prev->next=c; 
    client_list->prev=c; 
  } 
  client_count++; 
  update_gnome_client_list();
  reorder_dockicon_winlist(c);
 
  configure_client(c, CWHeight|CWWidth|CWX|CWY|CWBorderWidth, &wc, False, True); 
//  redraw_client_frame(c); 
 
  /* Set client to correct state */   
  if(attr.map_state==IsViewable) 
  { 
    if(get_wm_state(c->window)==IconicState) {
      raise_client(c);
      // turn new client into an iconbar
      minmax_client(c);
      lower_client(c);
      redraw_client_frame(c);
    } else {
      raise_client(c); }
  }
  else if(get_wm_state(c->window)==NormalState) 
    raise_client(c);
  else
    redraw_client_frame(c);
 
  /* Button grab for click-to-focus */ 
  XGrabButton(dpy, AnyButton, AnyModifier, c->frame, False, ButtonPressMask|ButtonReleaseMask, GrabModeSync, GrabModeAsync, None, None); 
 
#ifdef KEYBOARD_SUPPORT 
  grab_keys(c->window); 
#endif /* KEYBOARD */ 
 
#ifdef WORKSPACE_SUPPORT 
  set_gnome_hint(c->window, WIN_WORKSPACE, workspace); 
#endif /* WORKSPACE */ 
 
  /* Check if we want to focus new clients */ 
  if((c->is_icon==NOT_ICONBAR) && config.focus_new && !(get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS)) 
    set_focus_client(c); 

  if(showfront!=0 && gX11InFront==0)
    Bring_To_Front(0);

// 'Jig' it by one pixel (to make sure content appears on secondary screens)...
/*  wc.x=c->x+1; 
  wc.y=c->y; 
  wc.height=c->height; 
  wc.width=c->width; 
  configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, False); 
 
  wc.x=c->x-1; 
  wc.y=c->y; 
  wc.height=c->height; 
  wc.width=c->width; 
  configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, False); 
*/
  XUngrabServer(dpy);

  sprintf(lastsub,"TRACE1: exit create_new_client\n");
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: exit create_new_client\n");
#endif
} 

/*
 * Initialise the flag for checking if we've done a client during a 'nearest iconbar' search
 */
 
void init_nearest_iconbar()
{
  Client *c=NULL;
  int i;
  
  c=client_list;
  for(i=0;i<client_count;i++) {
    c->done_icon=0;
    c=c->next;
  }
}

/*
 * Find the iconbar nearest (vertically only) to the supplied y co-ordinate (ypos)
 */
 
Client *get_nearest_iconbar(int ypos,int state)
{
  int i, diff, mindif;
  Client *c=NULL, *ac;
  
  mindif=10000;
  ac=client_list;
  for(i=0;i<client_count;i++) {
    if(ac->is_icon==state || ac->is_icon==ICONBAR_RAISED || state==NOT_ICONBAR) {
      // only want clients that have already been an iconbar
      if(ac->icon_y!=-10000) {
        diff=ac->icon_y-ypos;
        if(diff<0) diff=-diff;
        if(diff<=mindif) {
          // only take clients that have not already been found in a previous call
          if(ac->done_icon==0) {
            if(state==NOT_ICONBAR) {
              // want a list of iconbar positions for every window that has been 'iconbar-ised'
              c=ac;
              mindif=diff;
            } else if(state==ICONBAR_HIDDEN) {
              // only want iconbars that are hidden
              c=ac;
              mindif=diff;
            } else if(state==ICONBAR_VISIBLE) {
              // only want iconbars that are visible
              c=ac;
              mindif=diff;
            }
          }
        }
      }
    }
    ac=ac->next;
  }
  
  // Note this client as having been found
  if(c!=NULL)
    c->done_icon=1;
    
  return c;
}

/*
 * Hide/unhide a particular iconbar
 */

void toggle_slide_iconbar(Client *c)
{
  XWindowChanges wc;

  sprintf(lastsub,"TRACE1: toggle_slide_iconbar: %#lx\n",c);
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: toggle_slide_iconbar: %#lx\n",c);
#endif

  if(c==NULL)
    return;

  if(c->is_icon==ICONBAR_HIDDEN) {
#ifdef SHOW_WEIRD
    if(iconbar_state==ICONBAR_VISIBLE)
      printf("%s oddity: Iconbar is hidden, but state is visible!!\n",PROGNAME);
#endif
    if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
      wc.x=c->icon_x+c->icon_w-100;
    else
      wc.x=c->icon_x;
    wc.y=c->icon_y;
    wc.width=100;
    wc.height=c->icon_h;
    configure_client(c, CWWidth|CWX, &wc, True, False); 
    c->is_icon=ICONBAR_VISIBLE;
  } else if(c->is_icon==ICONBAR_VISIBLE) {
    if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
      wc.x=c->icon_x+c->icon_w-icontab_width;
    else
      wc.x=c->icon_x;
    wc.y=c->icon_y;
    wc.width=icontab_width;
    wc.height=c->icon_h;
    configure_client(c, CWWidth|CWX, &wc, True, False);
    c->is_icon=ICONBAR_HIDDEN; 
    wc.stack_mode=Below; 
    configure_client(c, CWStackMode, &wc, True, False);
  }
}

/*
 * Hide/unhide all iconbars
 */
 
void slide_all_iconbars(Client *c)
{
  XWindowChanges wc;
  int i, k, neary;
  Client *ac;
  
  sprintf(lastsub,"TRACE1: slide_all_iconbars (client: %#lx)\n",c);
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: slide_all_iconbars (client: %#lx)\n",c);
#endif

  lastIconbarOutTicks=0;
  
  if(c==NULL) {
    if(iconbar_state==ICONBAR_HIDDEN) {
#ifdef SHOW_WEIRD
      printf("%s oddity: Iconbars already hidden??\n",PROGNAME);
#endif
      return; }
    iconbar_state=ICONBAR_HIDDEN;
    neary=nexticon_y;
    for(k=icontab_width;k>=icontab_width;k-=SLIDESTEP) {
      init_nearest_iconbar();
      ac=get_nearest_iconbar(neary,ICONBAR_VISIBLE);
      while(ac!=NULL) {
        if(ac->is_icon==ICONBAR_VISIBLE || ac->is_icon==ICONBAR_RAISED) {
          if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
            wc.x=ac->icon_x+ac->icon_w-k;
          else
            wc.x=ac->icon_x;
          wc.y=ac->icon_y;
          wc.width=k;
          wc.height=ac->icon_h;
          configure_client(ac, CWWidth|CWX, &wc, True, False);
          wc.stack_mode=Below;
          configure_client(ac, CWStackMode, &wc, True, False);
        }
        ac=get_nearest_iconbar(neary,ICONBAR_VISIBLE);
      }
    }
    ac=client_list;
    for(i=0;i<client_count;i++) {
      if(ac->is_icon==ICONBAR_VISIBLE || ac->is_icon==ICONBAR_RAISED)
        ac->is_icon=ICONBAR_HIDDEN; 
      ac=ac->next;
    }
  } else {
    if(iconbar_state==ICONBAR_VISIBLE) {
//#ifdef SHOW_WEIRD
//      printf("Iconbars already visible\n");
//#endif
      return; }
    iconbar_state=ICONBAR_VISIBLE;
    wc.stack_mode=Above; 
    XConfigureWindow(dpy, check_win, CWStackMode, &wc);
    neary=c->icon_y;
    c->is_icon=ICONBAR_RAISED;
    for(k=100;k<101;k+=SLIDESTEP) {
      if(c->is_icon==ICONBAR_HIDDEN || c->is_icon==ICONBAR_RAISED) {
        wc.stack_mode=Above; 
        configure_client(c, CWStackMode, &wc, True, False);
        if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
          wc.x=c->icon_x+c->icon_w-k;
        else
          wc.x=c->icon_x;
        wc.y=c->icon_y;
        wc.width=k;
        wc.height=c->icon_h;
        configure_client(c, CWWidth|CWX, &wc, True, False);
      }
      init_nearest_iconbar();
      ac=get_nearest_iconbar(neary,ICONBAR_HIDDEN);
      while(ac!=NULL) {
        if(ac->is_icon==ICONBAR_HIDDEN || ac->is_icon==ICONBAR_RAISED) {
          wc.stack_mode=Above; 
          configure_client(ac, CWStackMode, &wc, True, False);
          if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
            wc.x=ac->icon_x+ac->icon_w-k;
          else
            wc.x=ac->icon_x;
          wc.y=ac->icon_y;
          wc.width=k;
          wc.height=ac->icon_h;
          configure_client(ac, CWWidth|CWX, &wc, True, False);
        }
        ac=get_nearest_iconbar(neary,ICONBAR_HIDDEN);
      }
    }
    ac=client_list;
    for(i=0;i<client_count;i++) {
      if(ac->is_icon==ICONBAR_HIDDEN || ac->is_icon==ICONBAR_RAISED)
        ac->is_icon=ICONBAR_VISIBLE; 
      ac=ac->next;
    }
    c->is_icon=ICONBAR_RAISED;
  }
}

void hide_all_iconbars()
{
  slide_all_iconbars(NULL);
}

/*  
 *  
 * Get client frame from window ID, from window frame ID, title window ID, or from OS X WindowRef
 * New option - send in win with negative value to avoid check on lastenter (should speed it up a bit)
 */ 

Client *get_client_from_window(Window w, int win) 
{ 
  Client *c=NULL; 
  Client *ct=NULL;
  Client *cf=NULL;
  Client *cw=NULL;
  long i;
  int docheck=1; 
 
  sprintf(lastsub,"TRACE3: get_client_from_window (w,win,client_count: %#lx,%d,%d)\n",w,win,client_count);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: get_client_from_window (w,win,client_count: %#lx,%d,%d)\n",w,win,client_count); 
#endif 

  if(w==0 || client_count==0 || w==root || w==check_win ) {
    sprintf(lastsub,"TRACE3: exiting get_client_from_window (got nothing!)\n"); 
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: exiting get_client_from_window (got nothing!)\n"); 
#endif 
    return NULL; }

  if(win<0) {
    win=-win;
    docheck=0; }
    
  c=client_list; 
//#ifdef DEBUG
//   printf("   checking client [%#lx] %#lx _%#lx_\n",c->frame,c->window,c->title);
//#endif

// Just putting the entire loop into these if blocks seems to have brought down the
// CPU time quite noticeably (according to 'top', anyway)...
    if(win==TITLE) 
    { 
      for(i=0;i<client_count;i++) 
      { 
      // No lastenter check here
      if(c->title==w) {
//        ct=c;
  sprintf(lastsub,"TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
#endif 
        return c; 
      }
      c=c->next; 
      } 
    } 
    else if(win==FRAME_LE) 
    { 
      for(i=0;i<client_count;i++) 
      { 
      if(c->frame==w) {
        cf=c;
        // May want to continue and do a lastenter check...
        if(w!=lastenter || docheck==0) {
  sprintf(lastsub,"TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
#endif 
          return c;  }
      }
      c=c->next; 
      } 
    }  
    else  if(win==OSXWINREF)
    { 
      for(i=0;i<client_count;i++) 
      { 
      // No lastenter check here
      if(c->osxwinfo.winref==(long) w) { 
//        cw=c;
  sprintf(lastsub,"TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
#endif 
        return c; 
      }
      c=c->next; 
      } 
    }
    // default is to look for the client that has the actual X window
    else 
    { 
      for(i=0;i<client_count;i++) 
      { 
      if(c->window==w) { 
//        cw=c;
        // May want to continue and do a lastenter check...
//        if(c->frame!=lastenter || docheck==0)
  sprintf(lastsub,"TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: exiting get_client_from_window (got: %#lx)\n",c); 
#endif 
          return c; 
      }
      c=c->next; 
      } 
    } 

// do a check for the sanity of lastenter
  if(w==lastenter) {
//    if(win==TITLE) {
//      if(ct==NULL) {
//#ifdef SHOW_WEIRD
//         printf("  failed to find title for lastenter - setting to zero\n");
//#endif
//         lastenter=0; }
//      return ct;
//    } else if(win==FRAME) {
      if(cf==NULL) {
#ifdef SHOW_WEIRD
         printf("%s oddity: failed to find frame for lastenter - setting to zero\n",PROGNAME);
#endif
         lastenter=0; }
//      return cf;
//    } else {
//      if(cw==NULL) {
//#ifdef SHOW_WEIRD
//         printf("  failed to find window for lastenter - setting to zero\n");
//#endif
//         lastenter=0; }
//    }
  }

  sprintf(lastsub,"TRACE3: leaving get_client_from_window (got: %#lx)\n",cf); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: leaving get_client_from_window (got: %#lx)\n",cf); 
#endif

  return cf; 
} 
 
/* 
 * 
 * Get client with focus 
 * 
 */ 

Client *get_focus_client() 
{ 
  Client *c; 
  Window focus_win; 
  int dummy; 
 
  sprintf(lastsub,"TRACE3: get_focus_client (lastfocus: %#lx)\n",lastfocus); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_focus_client (lastfocus: %#lx)\n",lastfocus); 
#endif 
 
  XGetInputFocus(dpy, &focus_win, &dummy); 
  c=get_client_from_window(focus_win, WINDOW);

#ifdef DEBUG
  if(c==NULL)
    printf("  got no client for focus_win %#lx (%#lx)\n",focus_win,lastfocus);
  else
    printf("  got focus_win: %#lx (%#lx) [%#lx] %#lx\n",focus_win,lastfocus,c->frame,c->window);
#endif

// hack the focus for when menus are shown from a window -oroborus used to make
// an X11 nedit window 'inactive' when a menu was pulled down from it
  hacking_focus=False;
//  if(gX11InFront==0) return;
  
  if(c!=NULL)
    lastfocus=focus_win;
  else if(focus_win==check_win)
    lastfocus=0;
  else if(lastfocus!=0) {
    c=get_client_from_window(lastfocus, WINDOW);
    if(c!=NULL)
      hacking_focus=True;
    else
      lastfocus=0;
  }

  sprintf(lastsub,"TRACE3: leaving get_focus_client (c: %#lx  hf: %d)\n",c,hacking_focus); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: leaving get_focus_client (c: %#lx  hf: %d)\n",c,hacking_focus); 
#endif 
 
  return c;   
} 
 
/* 
 * 
 * Get next client in client list 
 * 
 */ 

Client *get_next_client(Client *c) 
{ 
  long i; 
 
  sprintf(lastsub,"TRACE3: get_next_client\n"); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_next_client\n"); 
#endif 
 
  if(c!=NULL) 
  { 
    for(i=0;i<client_count;i++) 
    { 
      c=c->next; 
#ifdef WORKSPACE_SUPPORT 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS && get_gnome_hint(c->window, WIN_WORKSPACE)==workspace) 
#else 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS) 
#endif /* WORKSPACE */ 
        return c; 
    } 
  } 
  return NULL; 
} 

/* 
 * 
 * Get previous client in client list 
 * 
 */ 

Client *get_prev_client(Client *c) 
{ 
  long i; 
 
  sprintf(lastsub,"TRACE3: get_prev_client (client: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_prev_client (client: %#lx)\n",c); 
#endif 
 
  if(c!=NULL) 
  { 
    for(i=0;i<client_count;i++) 
    { 
      c=c->prev; 
#ifdef WORKSPACE_SUPPORT 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS && get_gnome_hint(c->window, WIN_WORKSPACE)==workspace) 
#else 
      if(!get_gnome_hint(c->window, WIN_HINTS)&WIN_HINTS_SKIP_FOCUS) 
#endif /* WORKSPACE */ 
        return c; 
    } 
  } 
  return NULL;
}
 
/* 
 * 
 * Get top-most client (and say whether iconbars are to be included in search)
 * 
 */ 

Client *get_top_most_client(int layer, int get_icons) 
{ 
  Window *windows, w1, w2; 
  int count, i; 
  Client *top=NULL; 
  Client *c=NULL; 
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE3: get_top_most_client (layer,get_icons: %d,%d)\n",layer,get_icons); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_top_most_client (layer,get_icons: %d,%d)\n",layer,get_icons); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
#ifdef LAYER_SUPPORT 
  for(i=0;i<count;i++) 
  { 
    c=get_client_from_window(windows[i], FRAME); 
    if(c!=NULL && (c->is_icon==NOT_ICONBAR || get_icons)) 
    { 
      XGetWindowAttributes(dpy, c->window, &attr); 
      if(attr.map_state==IsViewable || get_icons) 
      { 
        if(get_gnome_hint(c->window, WIN_LAYER)<=layer) 
          top=c; 
        else 
          break; 
      } 
    } 
  } 
#else 
  for(i=count-1;i>=0;i--) 
  { 
    c=get_client_from_window(windows[i], FRAME); 
    if(c!=NULL && (c->modal_state & ANY_FLOATING)==0 && (c->is_icon==NOT_ICONBAR || get_icons)) 
    { 
      XGetWindowAttributes(dpy, c->window, &attr); 
      if(attr.map_state==IsViewable) { 
        top=c;
        break; }
    } 
  } 
#endif /* LAYER */ 
  if(windows) XFree(windows);
  
  if(top==NULL)
    FocusOSXWin(NULL);
  else if( top->is_icon!=NOT_ICONBAR )
    FocusOSXWin(NULL);

  return top; 
} 

/* 
 * 
 * Get top_most client which contains point x,y (and say whether iconbars are to be included)
 * 
 */ 

Client *get_client_container(int x, int y, int get_icons)
{
  Window *windows, w1, w2; 
  int count, i; 
  Client *top=NULL; 
  Client *c=NULL; 
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE2: get_client_container (x,y,get_icons: %d,%d %d)\n",x,y,get_icons); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: get_client_container (x,y,get_icons: %d,%d %d)\n",x,y,get_icons); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
  
  for(i=count-1;i<=0;i--) 
  { 
    c=get_client_from_window(windows[i], FRAME); 
    if(c!=NULL && (c->is_icon==NOT_ICONBAR || get_icons)) 
    { 
      XGetWindowAttributes(dpy, c->window, &attr); 
      if(attr.map_state==IsViewable || get_icons) { 
        if(x>=attr.x)
          if(x<attr.x+attr.width)
            if(y>=attr.y)
              if(y<attr.y+attr.height) {
                top=c;
                break; }
      }
    } 
  } 

  if(windows) XFree(windows);
  
  return top; 
}

/* 
 * 
 * Get bottom-most client (and say whether iconbars are to be included)
 * 
 */ 

Client *get_bottom_most_client(int layer, int get_icons) 
{ 
  Window *windows=NULL, w1, w2; 
  int count, i; 
  Client *c=NULL; 
  Client *bot=NULL;
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE3: get_bottom_most_client (layer,get_icons: %d,%d)\n",layer,get_icons); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_bottom_most_client (layer,get_icons: %d,%d)\n",layer,get_icons); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
  for(i=0;i<count;i++) 
  { 
    c=get_client_from_window(windows[i], FRAME); 
    if(c!=NULL && (c->is_icon==NOT_ICONBAR || get_icons)) 
    { 
      XGetWindowAttributes(dpy, c->window, &attr); 
      if(attr.map_state==IsViewable || get_icons) 
      { 
#ifdef LAYER_SUPPORT 
        if(get_gnome_hint(c->window, WIN_LAYER)>=layer) { 
          bot=c;
          break; }
#else 
        bot=c;
        break; 
#endif /* LAYER */ 
      } 
    } 
  } 
  if(windows) XFree(windows);  
  return bot; 
} 

/*
 * A new window has been chosen from the dock icon's window list, so change the focus in X11
 */
 
void change_focus(long winref)
{
  Client *c=NULL;
  
  sprintf(lastsub,"TRACE1: change_focus (client: %#lx)\n",c);
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: change_focus (client: %#lx)\n",c);
#endif

  c=get_client_from_window((Window) winref,OSXWINREF);
  
  if(c!=NULL) {
    if(c->is_icon==NOT_ICONBAR) {
      raise_client(c);
      set_focus_client(c);
    } else {
      c->is_icon=ICONBAR_RAISED;
      raise_client(c);
      minmax_client(c);
    }
  }
}

/* 
 *  
 * Raise client window 
 * 
 */ 

void raise_client(Client *c) 
{ 
  Client *orgc;
  XWindowChanges wc; 

  sprintf(lastsub,"TRACE2: raise_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: raise_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
#endif 
 
  if(c==NULL) 
  { 
#ifdef SHOW_WEIRD 
    printf("%s oddity: client is NULL in raise_client\n",PROGNAME); 
#endif 
    return;
  } 

  orgc=c;
  if(c->is_icon==NOT_ICONBAR)
    c=check_client_modality(orgc,1);
#ifdef DEBUG 
    printf("  raising client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
 
  if(c->is_icon==NOT_ICONBAR)
    show_client(c); 
  wc.stack_mode=Above; 
  XConfigureWindow(dpy, check_win, CWStackMode, &wc);
  wc.stack_mode=Above; 
  configure_client(c, CWStackMode, &wc, c->is_icon!=NOT_ICONBAR, False); 
} 

/*
 * Force a particular client to be raised, with no regard for modal children etc.
 */
 
void force_raise_client(Client *c) 
{ 
//  Client *orgc;
  XWindowChanges wc; 
 
  sprintf(lastsub,"TRACE3: force_raise_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: force_raise_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
#endif 
 
  if(c==NULL) 
  { 
#ifdef SHOW_WEIRD 
    printf("%s oddity: client is NULL in force_raise_client\n",PROGNAME); 
#endif 
    return;
  } 

#ifdef DEBUG 
    printf("  force-raising client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
 
  if(c->is_icon==NOT_ICONBAR)
    show_client(c);
  else {
    minmax_client(c);
  }
  
  wc.stack_mode=Above; 
  XConfigureWindow(dpy, c->frame, CWStackMode, &wc);
//  wc.stack_mode=Above; 
//  configure_client(c, CWStackMode, &wc, c->is_icon!=0); 
} 
 
void raise_all_clients()
{
  Window *windows=NULL, w1, w2; 
  int count, i; 
  Client *c=NULL; 
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE1: raise_all_clients\n"); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: raise_all_clients\n"); 
#endif 
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
  for(i=0;i<count;i++) 
  { 
    c=get_client_from_window(windows[i], FRAME); 
    if(c!=NULL && (c->is_icon==NOT_ICONBAR)) 
    { 
      XGetWindowAttributes(dpy, c->window, &attr); 
      if(attr.map_state==IsViewable) 
        force_raise_client(c);
    } 
  } 
  if(windows) XFree(windows);  
}

/*
 *
 * Entered/Left a client
 *
 */

void entered_client(Client *c, int can_focus)
{
  Client *fc;
  
#ifdef SHOW_ENTER_LEAVE
  printf("entered_client [%#lx] %#lx (%#lx , %#lx)\n",c==NULL?0:c->frame,c==NULL?0:c->window,lastfocus,overicon);
#endif
/*
  if(overicon==NULL && iconbar_state==ICONBAR_VISIBLE) {
#ifdef SHOW_WEIRD
    printf("Overicon is NULL, but iconbar visible? %d\n",iconbar_state);
#endif
    if(c==NULL)
      will_hide_all_iconbars();
    else if(c->is_icon!=NOT_ICONBAR)
      overicon=c;
    else
      overicon=client_list;
  }
*/  
  if(overicon!=NULL) {
// we were over an iconbar, but the bars were not visible??
#ifdef SHOW_WEIRD
     if(iconbar_state!=ICONBAR_VISIBLE)
       printf("%s oddity: Overicon: %#lx , but not visible? %d\n",PROGNAME,overicon,iconbar_state);
#endif
     if(overicon->is_icon!=ICONBAR_RAISED) {
// we were over an iconbar, but it was not lit up??
#ifdef SHOW_WEIRD
       printf("%s oddity: Overicon: %#lx , but not raised? %d\n",PROGNAME,overicon,overicon->is_icon);
#endif
       overicon=NULL;
// if we're not over a new one then hide the iconbars
       if(c==NULL)
         will_hide_all_iconbars();
       else if(c->is_icon==NOT_ICONBAR)
         will_hide_all_iconbars();
     } else if(overicon!=c) {
// over a different iconbar (or none) to what we were
       if(iconbar_state!=ICONBAR_VISIBLE) {
#ifdef SHOW_WEIRD
          printf("%s oddity: How did that happen? iconbar_state=%d, overicon=%#lx\n",PROGNAME,iconbar_state,overicon);
#endif
          slide_all_iconbars(overicon);
       }
       if(c==NULL) {
// not over a new iconbar, so prepare to hide them
         overicon->is_icon=ICONBAR_VISIBLE;
         will_hide_all_iconbars();
         if( overicon->icon_w!=100 )
           resize_client(overicon,100,overicon->icon_h,True);
         else
           redraw_client_frame(overicon);
//         if( overicon->icon_w!=icontab_width )
//           resize_client(overicon,icontab_width,overicon->icon_h,True);
       } else if(c->is_icon==NOT_ICONBAR) {
// not over a new iconbar, so prepare to hide them
         overicon->is_icon=ICONBAR_VISIBLE;
         will_hide_all_iconbars();
         if( overicon->icon_w!=100 )
           resize_client(overicon,100,overicon->icon_h,True);
         else
           redraw_client_frame(overicon);
//         if( overicon->icon_w!=icontab_width )
//           resize_client(overicon,icontab_width,overicon->icon_h,True);
       } else {
// still over an iconbar, so don't hide them, but make sure last one is redrawn correctly
         overicon->is_icon=ICONBAR_VISIBLE;
         if( overicon->icon_w!=100 )
           resize_client(overicon,100,overicon->icon_h,True);
         else
           redraw_client_frame(overicon);
       }
//       lower_client(overicon);
       overicon=NULL;
     }
  } else if(c!=NULL) {
     if(iconbar_state!=ICONBAR_HIDDEN) {
//#ifdef SHOW_WEIRD
//       printf("Iconbar not hidden, but overicon NULL?!\n");
//#endif
       will_hide_all_iconbars();
     }
  }

  if(c==NULL)
  {
    fc=get_focus_client();
    if(config.focus_style==FOCUS_SLOPPY && can_focus)
      set_focus_client(NULL);
    else if(fc!=NULL) {
      if((fc->auto_properties & AUTO_FOCUS)!=0) {
        fc=get_top_most_client(WIN_LAYER_NORMAL,False);
        set_focus_client(fc);
      }
    }
  }
  else if(c->is_icon!=NOT_ICONBAR)
  {
// we are over a new iconbar, show them if not already
    slide_all_iconbars(c);
// prepare for new iconbar by changing width (if necessary) and 'light it up'
    c->is_icon=ICONBAR_RAISED;
    if( c->icontxt_w>=TW-6 )
      resize_client(c,c->icontxt_w+6-LW-RW+TLW+TRW,c->icon_h,True);
    redraw_client_frame(c);
    raise_client(c);
    overicon=c;
  }
  else if(can_focus) {
    if(config.focus_style==FOCUS_SLOPPY || (c->auto_properties & AUTO_FOCUS)==AUTO_FOCUS)
      if(gX11InFront==1) set_focus_client(c);
  }

}

/*
 * Glue routine to be called from main.c to say we've left a client
 */
void left_client()
{
    entered_client(NULL,0);
}

/*
 * Want to check if a client has modal children, and returns the 'newest' child
 * Also, raises modal windows as it findds them if raiseit is non-zero
 */
 
Client *check_client_modality(Client *orgc, int raiseit)
{
Client *newc,*c,*tmpc;

  c=orgc;
  newc=c;
  if(c->first_child) {
    if(c->modal_state&HAS_ANY_MODAL_CHILD) {
      if( raiseit ) force_raise_client(c);
      c=c->first_child;
      if(c->modal_state&ANY_MODAL) {
        newc=c;
        if( raiseit ) force_raise_client(c);
      }
      if(c->modal_state&HAS_ANY_MODAL_CHILD)
        newc=check_client_modality(c,raiseit);
      while(c->next_sibling) {
        c=c->next_sibling; 
        if(c->modal_state&ANY_MODAL) {
          newc=c;
          if( raiseit ) force_raise_client(c);
        }
        if(c->modal_state&HAS_ANY_MODAL_CHILD)
          newc=check_client_modality(c,raiseit);
      }
    }
  }
  return newc;
}

/* 
 *  
 * Give client window keyboard focus 
 * 
 */ 

void set_focus_client(Client *c) 
{ 
  Client *orgc;
  int can_af=-1, can_ar=-1;
  
  sprintf(lastsub,"TRACE1: set_focus_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: set_focus_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
#endif 
 
  if(c==NULL) 
  { 
#ifdef DEBUG
    printf("  got NULL client in set_focus_client\n"); 
#endif
    XSetInputFocus(dpy, check_win, RevertToNone, CurrentTime); 
    lastfocus=0;
    Setup_Window_Menu(0, 0, 0, 0, 0, 0, 0);
    return;
  } 

  orgc=c;
  c=check_client_modality(orgc,0);

  if(c->is_icon==NOT_ICONBAR)
  { 
#ifdef DEBUG 
    printf("  focusing client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
    XSetInputFocus(dpy, c->window, RevertToNone, CurrentTime);
    if(lastfocus!=c->window) {
       lastfocus=c->window;
       redraw_client_frame(c);
    }
    if((c->auto_properties & AUTO_RAISE)!=0) {
       if((c->auto_properties & AUTO_FOCUS)!=0)
         can_af=1;
       raise_client(c);
       can_ar=1;
    } else if((c->auto_properties & AUTO_FOCUS)==0 && config.focus_style==FOCUS_CLICK)
       raise_client(c);
    else if((c->auto_properties & AUTO_FOCUS)!=0)
       can_af=1;
    if(c->osxwinfo.winref!=0) {
       FocusOSXWin(&(c->osxwinfo));
       Setup_Window_Menu(c->has_frame, 1, 1, c->can_max, 0, can_af, can_ar);
    }
  } 
} 

/*
 * When XDarwin is brought in front, we have to also (unfortunately) bring all
 * clients that have auto-focus set (to work around mouse tracking issue in XDarwin) 
 */
void raise_auto_clients(Client *skipc)
{
  Client *c;
  int i;
  
  sprintf(lastsub,"TRACE2: raise_auto_clients \n");
  subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: raise_auto_clients \n");
#endif

  c=client_list;
  if(c==NULL) return;
  
  for(i=0;i<client_count;i++) {
    if(c->is_icon==NOT_ICONBAR)
      if((c->auto_properties & AUTO_FOCUS)!=0)
        if(c!=skipc)
          raise_client(c);
    c=c->next;
  }
}

/*
 * OroborOSX has come in front, so reactivate the focus window
 */
void come_to_front()
{
    Client *c=NULL, *ctop=NULL;

  sprintf(lastsub,"TRACE2: come_to_front \n");
  subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: come_to_front \n");
#endif

    c=get_focus_client();
    if( c )
      redraw_client_frame(c);
    
    ctop=get_top_most_client(WIN_LAYER_NORMAL,False);
    raise_auto_clients(ctop);
    if( ctop )
      raise_client(ctop);
      
#ifdef SHOW_EVENTS
    printf("come_to_front (focus,top: %#lx,%#lx)\n",c,ctop);
#endif
}

/*
 * Another OS X app has come in front, so deactivate the focus window
 */

void gone_to_back()
{
    Client *c=NULL;
    
  sprintf(lastsub,"TRACE2: gone_to_back \n");
  subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE2: gone_to_back \n");
#endif

    c=get_focus_client();
    if( c )
      redraw_client_frame(c);
    XUngrabPointer(dpy, CurrentTime); 
    
#ifdef SHOW_EVENTS
    printf("gone_to_back (focus-client: %#lx)\n",c);
#endif
}

/* 
 * 
 * Lower client window 
 * 
 */ 

void lower_client(Client *c) 
{ 
  XWindowChanges wc; 
 
  sprintf(lastsub,"TRACE3: lower_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: lower_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
#endif 
 
  if(c==NULL) {
#ifdef SHOW_WEIRD
    printf("%s oddity: client is NULL in lower_client\n",PROGNAME);
#endif
    return;
  }

  if((c->auto_properties & (AUTO_FOCUS|AUTO_RAISE))!=0) {
//#ifdef SHOW_WEIRD
    printf("  %s question... should an auto-focus/auto-raise client really be lowered?  -hmmmm...\n",PROGNAME);
//#endif
//    return;
  }
  
#ifdef DEBUG
  printf("  lowering client [%#lx] %#lx\n",c->frame,c->window);
#endif

  wc.stack_mode=Below; 
  configure_client(c, CWStackMode, &wc, c->is_icon!=NOT_ICONBAR, False); 
} 
 
/* 
 *  
 * Show client window 
 * 
 */
 
void show_client(Client *c) 
{ 
  sprintf(lastsub,"TRACE3: show_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: show_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
#endif 
 
  if(c==NULL) {
#ifdef SHOW_WEIRD
    printf("%s oddity: client is NULL in show_client\n",PROGNAME);
#endif
    return;
  } 

#ifdef DEBUG
  printf("  showing client [%#lx] %#lx\n",c->frame,c->window);
#endif
  c->is_icon=NOT_ICONBAR;
  XMapWindow(dpy, c->frame); 
  XMapWindow(dpy, c->window); 
  set_wm_state(c->window, NormalState); 
} 
 
/* 
 * 
 * Hide client window - this is the original iconisation from oroborus which is no longer used
 * 
 */ 

void old_hide_client(Client *c) 
{ 
  sprintf(lastsub,"TRACE3: hide_client (old)\n"); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: hide_client (old)\n"); 
#endif 

  if(c==NULL) {
#ifdef SHOW_WEIRD
    printf("%s oddity: client is NULL in old_hide_client\n",PROGNAME);
#endif
    return;
  }

#ifdef DEBUG 
  printf("  hiding client [%#lx] %#lx\n",c->frame,c->window);
#endif

  /* We need to ignore one of the unmaps, otherwise our client gets deleted */ 
  c->ignore_unmap++; 
  XUnmapWindow(dpy, c->frame); 
  XUnmapWindow(dpy, c->window); 
  set_wm_state(c->window, IconicState); 
} 
 
/*  
 * 
 * Configure client - the main resizing, raising, lowering function 
 * 
 */ 

void configure_client(Client *c, unsigned int mask, XWindowChanges *wc, int frame_only, int config_win) 
{ 
  XConfigureEvent ce; 
  int w, w1, h, h1, orig_is_icon;
  unsigned int orgmask;
 
  sprintf(lastsub,"TRACE2: configure_client (client,mask,frame_only,lastfocus: %#lx,%d,%d,%#lx)\n",c,mask,frame_only,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: configure_client (client,mask,frame_only,lastfocus: %#lx,%d,%d,%#lx)\n",c,mask,frame_only,lastfocus); 
#endif 

  if(c==NULL) {
#ifdef SHOW_WEIRD
  printf("%s oddity: client is NULL in configure_client\n",PROGNAME);
#endif
    return;
  }

#ifdef DEBUG
  printf("  configuring client [%#lx] %#lx\n",c->frame,c->window);
#endif

  orig_is_icon=c->is_icon;
  
  if(!frame_only) {
// if it wants the whole frame, then it should not be an iconbar!
#ifdef SHOW_WEIRD
    if(c->is_icon!=NOT_ICONBAR)
      printf("%s oddity: Asked to configure an iconbar with whole frame!\n",PROGNAME);
#endif
    c->is_icon=NOT_ICONBAR;
    }

// choose the height and width depending on whether the client is an iconbar or not
  if(c->is_icon!=NOT_ICONBAR) {
    h=c->icon_h;
    w=c->icon_w;
    if(mask & CWX) c->icon_x=wc->x;
    if(mask & CWY) c->icon_y=wc->y;
  } else {
    h=c->height;
    w=c->width;
    if(mask & CWX) c->x=wc->x;
    if(mask & CWY) c->y=wc->y;
  }

  if(mask & CWWidth) 
  { 
    w=wc->width; 
 
    /*printf("Setting width: %d , %d , %d , %d , %d\n",w,c->icontxt_x,c->name_w,c->width,c->icon_w);*/

    if( c->is_icon!=NOT_ICONBAR && frame_only) {
      // reset the iconbar width
      if(w<1) w=1; 
      c->icon_w=w;
    } else {

      /* Get resize increments */ 
      if(c->size->flags & PResizeInc) 
      { 
        w1=(w-c->size->min_width)/c->size->width_inc; 
        w=c->size->min_width+(w1*c->size->width_inc); 
      } 
      /* Get min size */ 
      if(c->size->flags & PMinSize) 
        if(w<c->size->min_width) w=c->size->min_width; 
      /* Get max size */ 
      if(c->size->flags & PMaxSize) 
        if(w>c->size->max_width) w=c->size->max_width; 
 
      /* Make sure we never set width to <1 */ 
      if(w<1) w=1; 

      c->width=w;
    }
  } 

  if(mask & CWHeight) 
  { 
    h=wc->height; 
 
    if( c->is_icon!=NOT_ICONBAR && frame_only ) {
      // reset the iconbar height
      if(h<1) h=1; 
      c->icon_h=h;
    } else {

      /* Get resize increments */ 
      if(c->size->flags & PResizeInc) 
      { 
        h1=(h-c->size->min_height)/c->size->height_inc; 
        h=c->size->min_height+(h1*c->size->height_inc); 
      } 
      /* Get min size */ 
      if(c->size->flags & PMinSize) 
        if(h<c->size->min_height) h=c->size->min_height; 
      /* Get max size */ 
      if(c->size->flags & PMaxSize) 
        if(h>c->size->max_height) h=c->size->max_height; 
 
      /* Make sure we never set height to <1 */ 
      if(h<1) h=1; 

      c->height=h;
    }
  } 
 
#ifdef LAYER_SUPPORT 
  /* Deal with window layers */ 
  if(mask & CWStackMode) 
  { 
    Client *sibling=NULL; 
 
    /* Takes care of layers */ 
    switch(wc->stack_mode) 
    { 
    case Above: 
      /* Window wants to be raised to top of its layer */ 
      sibling=get_top_most_client(get_gnome_hint(c->window, WIN_LAYER),True); 
      if(sibling==NULL) 
        wc->stack_mode=Below; 
      break; 
    case Below: 
      /* Window wants to be lowered to bottom of its layer */ 
      sibling=get_bottom_most_client(get_gnome_hint(c->window, WIN_LAYER),True); 
      break; 
    } 
    if(sibling==NULL) 
    { 
      if(mask & CWSibling) mask=mask ^ CWSibling; 
    } 
    else if(sibling==c) 
    { 
      if(mask & CWStackMode) mask=mask ^ CWStackMode; 
      if(mask & CWSibling) mask=mask ^ CWSibling; 
    } 
    else 
    { 
      mask=mask | CWSibling; 
      wc->sibling=sibling->frame; 
    } 
#ifdef WORKSPACE_SUPPORT 
    set_gnome_hint(c->window, WIN_WORKSPACE, workspace); 
#endif /* WORKSPACE */ 
  } 
#endif /* LAYER */ 
 
  orgmask=mask;
  
/* We don't want any stacking order changes on our actual client */ 
  if(mask & CWStackMode) mask=mask ^ CWStackMode; 
  if(mask & CWSibling) mask=mask ^ CWSibling;

//  if(!frame_only && mouse_mode==MOUSE_NORMAL)
  if(!frame_only)
  {
    if(config_win) {
      wc->x=LW;
      wc->y=TH;
      wc->width=c->width;
      wc->height=c->height;
      XConfigureWindow(dpy, c->window, mask, wc);
    } else
      XMoveResizeWindow(dpy,c->window, LW, TH, c->width, c->height);
  }

#ifdef SHAPE_SUPPORT 
  if((mask & (CWHeight|CWWidth))!=0 && !frame_only)
    set_client_shape(c,1);
#endif /* SHAPE */ 

  wc->x=FX; 
  wc->y=FY; 
  wc->width=w+LW+RW;
  wc->height=FH; 
  XConfigureWindow(dpy, c->frame, mask, wc);

  mask=orgmask;

// Now deal with stacking order changes
  if(mask & CWStackMode)
  {
    if(mask & CWSibling)
      XConfigureWindow(dpy, c->frame, CWStackMode | CWSibling, wc);
    else
      XConfigureWindow(dpy, c->frame, CWStackMode, wc);
  }
  
  /* If size has changed */ 
  if(mask & (CWHeight|CWWidth)) 
  { 
    if( c->has_frame || (c->is_icon!=NOT_ICONBAR) ) 
    { 
      /* Reposition and resize frame parts */ 

      if(c->is_icon!=NOT_ICONBAR && TW<1) {
        if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT) {
           if(c->top_left) {
              XMoveResizeWindow(dpy, c->top_left, TLX, TLY, icontab_width, TLH);
              if(c->top_right) XMoveResizeWindow(dpy, c->top_right, TRX+TRW-1, TRY, 1, TRH);
              if(c->title) XMoveResizeWindow(dpy, c->title, 1, TY, 1, TH); 
           } else if(c->top_right) {
              XMoveResizeWindow(dpy, c->top_right, TRX, TRY, icontab_width, TRH);
              if(c->top_left) XMoveResizeWindow(dpy, c->top_left, TLX, TLY, 1, TLH);
              if(c->title) XMoveResizeWindow(dpy, c->title, 1, TY, 1, TH); 
           }
        } else {
           if(c->top_right) {
              XMoveResizeWindow(dpy, c->top_right, TRX, TRY, TRW, TRH);
              if(c->top_left) XMoveResizeWindow(dpy, c->top_left, TLX, TLY, 1, TLH);
              if(c->title) XMoveResizeWindow(dpy, c->title, 1, TY, 1, TH); 
           } else if(c->top_left) {
              XMoveResizeWindow(dpy, c->top_left, TLX, TLY, TLW, TLH);
              if(c->title) XMoveResizeWindow(dpy, c->title, 1, TY, 1, TH); 
           }
        }
      } else {
        if(c->title) XMoveResizeWindow(dpy, c->title, TX, TY, TW<1?1:TW, TH); 
        if(c->top_left) XMoveResizeWindow(dpy, c->top_left, TLX, TLY, TLW, TLH); 
        if(c->top_right) XMoveResizeWindow(dpy, c->top_right, TRX, TRY, TRW, TRH);
      }
//      XMoveResizeWindow(dpy, c->top_right, w+LW+RW-TRW, TRY, TRW, TRH);

      if( c->is_icon!=NOT_ICONBAR ) {
        // window buttons need to be nowehere in sight for iconbars! (These are unmapped now...)
//        XMoveResizeWindow(dpy, c->close, -100, -100, CW, CH);
//        XMoveResizeWindow(dpy, c->icon, -100, -100, IW, IH);
//        XMoveResizeWindow(dpy, c->hide, -100, -100, HW, HH);
      } else if( frame_only) {
        // only doing frame...
        if(c->close) XMoveResizeWindow(dpy, c->close, CX, CY, CW, CH);
        if(c->icon) XMoveResizeWindow(dpy, c->icon, IX, IY, IW, IH);
        if(c->hide) XMoveResizeWindow(dpy, c->hide, HX, HY, HW, HH);
      } else {
        // doing everything...
        if(c->left) XMoveResizeWindow(dpy, c->left, LX, LY, LW, LH);
        if(c->right) XMoveResizeWindow(dpy, c->right, RX, RY, RW, RH);
        if(c->bottom_left) XMoveResizeWindow(dpy, c->bottom_left, BLX, BLY, BLW, BLH);
        if(c->bottom) XMoveResizeWindow(dpy, c->bottom, BX, BY, BW, BH);
        if(c->resize) XMoveResizeWindow(dpy, c->resize, RSX, RSY, RSW, RSH);
        if(c->close) XMoveResizeWindow(dpy, c->close, CX, CY, CW, CH);
        if(c->icon) XMoveResizeWindow(dpy, c->icon, IX, IY, IW, IH);
        if(c->hide) XMoveResizeWindow(dpy, c->hide, HX, HY, HW, HH);
      }
      redraw_client_frame(c);
    }

//    XConfigureWindow(dpy, c->frame, mask, wc); 

//    if( c->has_frame || (c->is_icon!=NOT_ICONBAR) ) 
//      redraw_client_frame(c);

    if(c->osxwinfo.winref!=0)
        SetOSXWinPos(&(c->osxwinfo),FX,FY,FW,FH,CX,CY,HX+HW,HY+HH);
        
//#ifdef SHAPE_SUPPORT 
//    set_client_shape(c);
//#endif /* SHAPE */ 
  }
//  else
//    XConfigureWindow(dpy, c->frame, mask, wc); 
 
  /* Notify client of window changes if changes to position or size */ 
//  if(mouse_mode==MOUSE_NORMAL && (mask & (CWX|CWY|CWHeight|CWWidth))!=0 && !frame_only && (c->is_icon==NOT_ICONBAR) ) 
  if((mask & (CWX|CWY|CWHeight|CWWidth))!=0 && !frame_only && (c->is_icon==NOT_ICONBAR) ) 
  { 
    ce.type=ConfigureNotify;
    ce.display=dpy;
    ce.event=c->window; 
    ce.window=c->window; 
    ce.x=c->x; 
    ce.y=c->y; 
    ce.width=c->width; 
    ce.height=c->height; 
    ce.border_width=0; 
    ce.above=c->frame; 
    ce.override_redirect=0; 
    XSendEvent(dpy, c->window, False, StructureNotifyMask, (XEvent *)&ce);
  }
  
#ifdef SHAPE_SUPPORT 
  if((mask & (CWHeight|CWWidth))!=0)
    set_client_shape(c,0);
#endif /* SHAPE */ 

//  c->is_icon=orig_is_icon;
} 
 
/* 
 * 
 * Move client to new position 
 * 
 */ 

void move_client(Client *c, int x, int y) 
{ 
  XWindowChanges wc; 
 
  sprintf(lastsub,"TRACE2: move_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: move_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
#endif
#ifdef DEBUG
  printf("  moving client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
 
  wc.x=x; 
  wc.y=y; 
  configure_client(c, CWX|CWY, &wc, c->is_icon!=NOT_ICONBAR, False); 
} 
 
/* 
 * 
 * Resize client 
 * 
 */ 

void resize_client(Client *c, int w, int h, int frame_only) 
{ 
  XWindowChanges wc;
  int mask;
 
  sprintf(lastsub,"TRACE2: resize_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: resize_client (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
#endif
#ifdef DEBUG
  printf("  resizing client [%#lx] %#lx\n",c->frame,c->window); 
#endif 
 
  wc.width=w; 
  wc.height=h;
  mask = CWWidth|CWHeight;
  
  if(c->is_icon!=NOT_ICONBAR) {
    // watch out for iconbars on the right!
    if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT) {
      mask=mask|CWX;
      wc.x=c->icon_x+c->icon_w-w;
    }
  }
  
  configure_client(c, mask, &wc, frame_only || (c->is_icon!=NOT_ICONBAR), False); 
} 
 
/* 
 * 
 * Toggle client maximized 
 * 
 */ 

void toggle_client_maximized(Client *c) 
{ 
  XWindowChanges wc; 
 
  if( c->is_icon!=NOT_ICONBAR )
    return;

  sprintf(lastsub,"TRACE1: toggle_client_maximized (client,can_max: %#lx,%d)\n",c,(c==NULL)?0:c->can_max); 
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: toggle_client_maximized (client,can_max: %#lx,%d)\n",c,(c==NULL)?0:c->can_max); 
#endif 
 
  if(c->can_max) 
  { 
    if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_MAXIMIZED) 
    { 
      c->x=c->old_x; 
      c->y=c->old_y; 
      c->width=c->old_width; 
      c->height=c->old_height; 
    } 
    else 
    { 
      int head=0; 
 
#ifdef XINERAMA_SUPPORT 
      if(xinerama)
        head=find_head(c->x+c->width/2, c->y+c->height/2); 
#endif /* XINERAMA */ 
 
      c->old_x=c->x; 
      c->old_y=c->y; 
      c->old_width=c->width; 
      c->old_height=c->height; 
      c->x=screen_x(head)+LW+config.left_margin; 
      c->y=screen_y(head)+TH+config.top_margin; 
      c->width=screen_width(head)-LW-RW-config.left_margin-config.right_margin; 
      c->height=screen_height(head)-TH-BH-config.top_margin-config.bottom_margin;
    } 
    wc.x=c->x; 
    wc.y=c->y; 
    wc.height=c->height; 
    wc.width=c->width; 
    configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, c->is_icon!=NOT_ICONBAR, False); 
    set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_MAXIMIZED); 
  } 
} 
 
/* 
 *  
 * Toggle client shaded 
 * 
 */ 

void toggle_client_shaded(Client *c) 
{ 
  int head;
  XWindowChanges wc;
  int mask;

  if(c->is_icon!=NOT_ICONBAR)
    return;

  sprintf(lastsub,"TRACE1: toggle_client_shaded (client: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: toggle_client_shaded (client: %#lx)\n",c); 
#endif 
 
  if(c->has_frame) 
  { 
    set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_SHADED); 

    wc.width=c->width; 
    wc.height=c->height;
    mask = CWWidth|CWHeight;
  
    if(c->is_icon!=NOT_ICONBAR) {
      // watch out for iconbars on the right!
      if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT) {
        mask=mask|CWX;
        wc.x=c->icon_x+c->icon_w-c->width;
      }
    }
  
    configure_client(c, mask, &wc, (get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)==WIN_STATE_SHADED, True); 
//    resize_client(c, c->width, c->height, True);

// Don't want the titlebar to be off the screen!
    if( xinerama ) {
       head=find_head(c->x,c->y);
       if(c->y-TH<screen_y(head))
         move_client(c, c->x, screen_y(head)+TH);
    } else {
       if(c->y-TH<0)
         move_client(c, c->x, TH );
    }
    
    redraw_client_frame(c); 
  }

}

/*
 * This has turned into a bit of a monster routine to toggle a client into/from an iconbar!
 */
 
void minmax_client(Client *c) 
{ 
  XWindowChanges wc;
  int i, nowt, nowx, nowy, noww, endt, endx, endy, endw;
  int shaded=0;
  Client *newc=NULL, *ac=NULL;
  Window ptrwin;

  sprintf(lastsub,"TRACE1: minmax_client (client: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: minmax_client (client: %#lx)\n",c); 
#endif 
 
    newc=get_focus_client();

    if( c->is_icon!=NOT_ICONBAR ) {
       if(gX11InFront==0)
         Bring_X_To_Front(1);
       if(newc) {
         set_focus_client(NULL);
         redraw_client_frame(newc);
         newc=NULL; }
       if(iconbar_state==ICONBAR_HIDDEN && c->is_icon==ICONBAR_HIDDEN)
         toggle_slide_iconbar(c);
       nowx=c->icon_x;
       nowy=c->icon_y;
       endx=c->x;
       endy=c->y;
       nowt=c->icontxt_x;
       noww=c->icon_w;
       if(shrinking_iconbars!=0) {
         endw=c->width;
         endt=c->name_x;
       } else {
         endt=nowt;
         endw=noww; }
       c->is_icon=ICONBAR_RAISED;
       // move iconbar from its location to where the window should be, changing width as necessary
       for(i=iconbar_steps-1;i>=0;i--) {
         wc.x=endx-(endx-nowx)*i/iconbar_steps;
         wc.y=endy-(endy-nowy)*i/iconbar_steps;
         wc.width=endw-(endw-noww)*i/iconbar_steps;
         wc.height=c->icon_h;
         c->icontxt_x=endt-(endt-nowt)*i/iconbar_steps;
         if(i==iconbar_steps-1)
           configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, True, False);
         else if(shrinking_iconbars!=0)
           configure_client(c, CWWidth|CWX|CWY, &wc, True, False);
         else
           configure_client(c, CWX|CWY, &wc, True, False); 
         //sleep(1);  -heh! interesting to watch it do this slowly...!
       }
       wc.x=c->x;
       wc.y=c->y;
       wc.width=c->width;
       wc.height=c->height;
       c->is_icon=NOT_ICONBAR;
       c->icon_x=nowx;
       if(noww!=100)
         if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
           c->icon_x=nowx+noww-100;
       c->icon_y=nowy;
       c->icon_w=100;
       if(shrinking_iconbars!=0)
         c->icontxt_x=nowt;
       XMapWindow(dpy,c->window);
       if(c->has_frame==0) {
         XUnmapWindow(dpy,c->top_left);
         XUnmapWindow(dpy,c->top_right);
         XUnmapWindow(dpy,c->title);
       } else {
         XMapWindow(dpy,c->close);
         XMapWindow(dpy,c->icon);
         XMapWindow(dpy,c->hide);
       }
    } else {
       // Tell OroborOSX this window is being iconised - doesn't actually do anything, since we still
       // want it to appear in the dock icon's window list, but may be useful at some point...
       if(c->osxwinfo.winref!=0)
         HideOSXWin(&(c->osxwinfo));
       nowx=c->x;
       nowy=c->y;
       noww=c->width;
       nowt=c->name_x;
       if( newc==c ) {
         c->is_icon=ICONBAR_RAISED;
       } else {
         c->is_icon=ICONBAR_VISIBLE; }
       // need these sub-windows back, if there was no frame
       if(c->has_frame==0) {
         XMapWindow(dpy,c->top_left);
         XMapWindow(dpy,c->top_right);
         XMapWindow(dpy,c->title);
       } else {
         XUnmapWindow(dpy,c->close);
         XUnmapWindow(dpy,c->icon);
         XUnmapWindow(dpy,c->hide);
       }
       newc=NULL;
       // Check to see if this window has been 'iconbar-ised' before - if not, set new values
       if(c->icon_x==-10000) {
         c->icon_w=100;
         if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_BOTRIGHT)
           c->icon_x=nexticon_x-c->icon_w;
         else
           c->icon_x=nexticon_x;
         c->icon_y=nexticon_y;
         c->icon_h=config.title.height;
         c->icontxt_x=3;
         c->icontxt_y=c->name_y;
         c->icontxt_w=c->name_w;
         c->icontxt_h=c->name_h;
         // Which way are the iconbars going?
         if(iconbarpos==ICONBARPOS_TOPLEFT || iconbarpos==ICONBARPOS_TOPRIGHT)
           nexticon_y+=config.title.height;
         else if(iconbarpos==ICONBARPOS_BOTLEFT || iconbarpos==ICONBARPOS_BOTRIGHT)
           nexticon_y-=config.title.height;
       }
       endx=c->icon_x;
       endy=c->icon_y;
       endw=c->icon_w;
       endt=c->icontxt_x;
       if(shrinking_iconbars==0) {
         noww=endw;
         nowt=endt; }
       // get rid of the actual window content
       c->ignore_unmap++;
       XUnmapWindow(dpy,c->window);
       // move iconbar from the window to where the it's going, changing width as necessary
       for(i=0;i<iconbar_steps;i++) {
         wc.width=noww-(noww-endw)*i/iconbar_steps;
         wc.height=c->icon_h;
         wc.x=nowx-(nowx-endx)*i/iconbar_steps;
         wc.y=nowy-(nowy-endy)*i/iconbar_steps;
         c->icontxt_x=nowt-(nowt-endt)*i/iconbar_steps;
         if(i==0)
           configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, True, False);
         else if(shrinking_iconbars!=0)
           configure_client(c, CWWidth|CWX|CWY, &wc, True, False);
         else
           configure_client(c, CWX|CWY, &wc, True, False); 
         //sleep(1);
       }
       wc.width=endw;
       wc.height=c->icon_h;
       wc.x=endx;
       wc.y=endy;
       if(shrinking_iconbars!=0)
         c->icontxt_x=endt;
       c->is_icon=ICONBAR_VISIBLE;
   }

    //shaded=get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED;
    
    // always set the window to shaded -otherwise the window 'flickers' twice when
    // returning from an iconbar, for some reason...
    set_gnome_hint(c->window, WIN_STATE, WIN_STATE_SHADED);
    shaded=WIN_STATE_SHADED;

    if( c->is_icon==NOT_ICONBAR ) {
       overicon=NULL;
       will_hide_all_iconbars();
       raise_client(c);
       set_focus_client(c);
       if(gX11InFront==0)
         Bring_To_Front(0);
    }

    configure_client(c, CWHeight|CWWidth|CWX|CWY, &wc, ((shaded==WIN_STATE_SHADED) || (c->is_icon!=NOT_ICONBAR)), False); 

    redraw_client_frame(c);

    if(c->is_icon!=NOT_ICONBAR) {
//      lower_client(c);
//      printf("Finished iconbar-ing... %d\n",iconbar_state);
      newc=get_top_most_client(WIN_LAYER_NORMAL,False);
      if( newc==NULL ) {
        set_focus_client(NULL);
      } else if( newc->is_icon!=NOT_ICONBAR ) {
        set_focus_client(NULL);
      } else
        set_focus_client(newc);

      ptrwin=get_pointer_win(root);
      
      if(ptrwin==c->frame) {
//        printf("over new iconbar: %#lx\n",c);
        // We've ended up with the cursor over the new iconbar, so make sure iconbars don't disappear!
        if(iconbar_state==ICONBAR_HIDDEN)
          slide_all_iconbars(c);
        else
          lastIconbarOutTicks=0;
        // ...and that it lights up
        c->is_icon=ICONBAR_RAISED;
        // ...and the width expands to see the whole text, if needed
        if( c->icontxt_w>=TW-6 )
          resize_client(c,c->icontxt_w+6-LW-RW+TLW+TRW,c->icon_h,True);
        redraw_client_frame(c);
        raise_client(c);
        overicon=c;
      } else {
        ac=get_client_from_window(ptrwin,FRAME_LE);
//        printf("over client: %#lx , ",ac);
        if(ac!=NULL) {
          if(ac->is_icon==NOT_ICONBAR) {
//            printf("not an iconbar...\n");
            if(iconbar_state==ICONBAR_VISIBLE) {
              c->is_icon=ICONBAR_VISIBLE;
              slide_all_iconbars(NULL);
            } else {
              c->is_icon=ICONBAR_VISIBLE;
              toggle_slide_iconbar(c);
            }
          } else {
//            printf("is an iconbar...\n");
            // we've ended up with the cursor over another iconbar, so make sure iconbars don't disappear
            if(iconbar_state!=ICONBAR_VISIBLE) {
              c->is_icon=ICONBAR_VISIBLE;
              slide_all_iconbars(ac);
            } else
              lastIconbarOutTicks=0;
          }
        } else {
//          printf("iconbar_state: %d\n",iconbar_state);
          if(iconbar_state==ICONBAR_VISIBLE) {
            c->is_icon=ICONBAR_VISIBLE;
            slide_all_iconbars(NULL);
          } else {
            c->is_icon=ICONBAR_VISIBLE;
            toggle_slide_iconbar(c);
          }
        }
      }
      
      newc=get_top_most_client(WIN_LAYER_NORMAL,False);

    } else {
//      printf("Finished un-iconbar-ing...  \n");
      raise_client(c);
      set_focus_client(c);
      if( shaded==WIN_STATE_SHADED )
        toggle_client_shaded(c);
// don't need this code any more...?
//      ptrwin=get_pointer_win(root);
//      printf("ptrwin,frame: %#lx %#lx\n",ptrwin,c->frame);
//      if(ptrwin==c->frame) {
//        printf("iconbar_state: %d\n",iconbar_state);
//        will_hide_all_iconbars();
//      }
    }
} 
 
#ifdef LAYER_SUPPORT 
/* 
 * 
 * Set client layer 
 * 
 */ 

void set_client_layer(Client *c, int layer) 
{ 
  sprintf(lastsub,"TRACE2: set_client_layer (client,layer: %#lx,%d)\n",c,layer); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: set_client_layer (client,layer: %#lx,%d)\n",c,layer); 
#endif 
 
  set_gnome_hint(c->window, WIN_LAYER, layer); 
  if(get_wm_state(c->window)==NormalState) 
    if(c->is_icon==NOT_ICONBAR)
      raise_client(c); 
} 
 
/* 
 * 
 * Get client layer 
 * 
 */ 

int get_client_layer(Client *c) 
{ 
  sprintf(lastsub,"TRACE2: get client layer (client: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: get client layer (client: %#lx)\n",c); 
#endif 
 
  return get_gnome_hint(c->window, WIN_LAYER); 
} 
#endif /* LAYER */ 
 
/* 
 * 
 * Delete client 
 * 
 */ 

void delete_client(Client *c, int remap) 
{ 
  sprintf(lastsub,"TRACE1: delete_client (client [window],lastfocus: %#lx [%#lx],%#lx)\n",c,c->window,lastfocus);
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: delete_client (client [window],lastfocus: %#lx [%#lx],%#lx)\n",c,c->window,lastfocus);
#endif

#ifdef SHOW_WEIRD
  if(c==NULL)
    printf("%s oddity: got NULL client in delete_client!\n",PROGNAME);
#endif
  if(c==NULL) return;

#ifdef DEBUG
  printf("  deleting client [%#lx] %#lx\n",c->frame,c->window); 
#endif 

  XGrabServer(dpy);
  
// make sure it has the right state...
  if(c->is_icon!=NOT_ICONBAR && remap) {
    if((get_gnome_hint(c->window,WIN_STATE) | WIN_STATE_SHADED)==WIN_STATE_SHADED)
      set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_SHADED); 
    c->ignore_unmap++; 
    XUnmapWindow(dpy, c->frame); 
    XUnmapWindow(dpy, c->window); 
    set_wm_state(c->window, IconicState);
  }

// This get_window_parent produces a BadWindow error, so inhibit it
  inhibit_errors=True;
  /* Check if frame is still the clients parent */
  if(get_window_parent(c->window)==c->frame) 
  { 
    inhibit_errors=False;
    /* Reparent our client back to root if it still exists */ 
    gravitate_client(c, REMOVE_GRAVITY); 
    XReparentWindow(dpy, c->window, root, c->x, c->y);
    set_wm_state(c->window, WithdrawnState); 
//    if(c->is_icon==NOT_ICONBAR)
//      set_wm_state(c->window, WithdrawnState); 

#ifdef KEYBOARD_SUPPORT
    /* Remove any key grabs on this window */
    ungrab_keys(c->window);
#endif
 
    /* Remap is true if we are quitting OroborOSX */ 
    if(remap) 
      XMapWindow(dpy, c->window); 
    else 
      XUnmapWindow(dpy, c->window); 
  } else
    inhibit_errors=False;

#ifdef DEBUG
  printf("  checking lastenter: [%#lx], [%#lx]\n",c->frame,lastenter);
  if( c->frame==lastenter ) printf("  window was last to be entered: [%#lx]\n",lastenter);
  printf("  checking lastc: [%#lx], [%#lx]\n",c,lastc);
  if( c==lastc ) printf("  window was lastc: [%#lx]\n",lastc);
#endif
  if( c->frame==lastenter ) lastenter=0;
  if( c==lastc ) lastc=NULL;

  if(c->osxwinfo.winref!=0)
    DestroyOSXWindow(&(c->osxwinfo));
    
  if(c->parent)
    remove_sibling_client(c);
    
  if(c->icon_y==nexticon_y-config.title.height)
    nexticon_y-=config.title.height;
	 
  /* Get rid of our frame window */ 
  XDestroyWindow(dpy, c->frame); 
 
  /* Remove reference from our client list */ 
  if(client_count==1) 
    client_list=NULL; 
  else 
  { 
    c->next->prev = c->prev; 
    c->prev->next = c->next; 
    if(c==client_list) 
      client_list=client_list->next; 
  } 
 
  if(c->name) XFree(c->name); 
  if(c->clientcmd) XFree(c->clientcmd); 
  if(c->clientclass) XFree(c->clientclass); 
  if(c->size) XFree(c->size); 
  free(c); 
 
  client_count--; 
  update_gnome_client_list();

  XUngrabServer(dpy);

  sprintf(lastsub,"TRACE2: delete_client -client removed\n");
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: delete_client -client removed\n");
#endif
} 
 
/* 
 * 
 * Send close message to client window 
 * 
 */ 

void close_client(Client *c) 
{ 
  int i, n, found=False; 
  Atom *protocols=NULL; 
  XEvent e; 
 
  sprintf(lastsub,"TRACE1: close_client (client: %#lx)\n",c); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: close_client (client: %#lx)\n",c); 
#endif 
 
#ifdef SHOW_WEIRD
  if(c==NULL) printf("%s oddity: got NULL client in close_client!\n",PROGNAME);
#endif
  if(c==NULL) return;
  
  if(XGetWMProtocols(dpy, c->window, &protocols, &n)) 
  { 
    for(i=0; i<n; i++) 
      if(protocols[i]==icccm[WM_DELETE]) found=True; 
    if(protocols) XFree(protocols);  
  }
  
  if(found) 
  { 
#ifdef DEBUG
    printf("  sending delete event to client [%#lx] %#lx\n",c->frame,c->window);
#endif
    e.type=ClientMessage; 
    e.xclient.window=c->window; 
    e.xclient.message_type=icccm[WM_PROTOCOLS]; 
    e.xclient.format=32; 
    e.xclient.data.l[0]=icccm[WM_DELETE]; 
    e.xclient.data.l[1]=CurrentTime; 
    XSendEvent(dpy, c->window, False, NoEventMask, &e); 
  } 
  else  {
#ifdef DEBUG
    printf("  killing client [%#lx] %#lx\n",c->frame,c->window);
#endif
    XKillClient(dpy, c->window); 
  }
} 
 
/* 
 * 
 * Redraw client borders 
 * 
 */ 

void redraw_client_frame(Client *c) 
{ 
  int state=FRAME_INACTIVE, shaded=0; 
  
  sprintf(lastsub,"TRACE3: redraw_client_frame (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: redraw_client_frame (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
#endif 

#ifdef SHOW_WEIRD
  if(c==NULL) printf("%s oddity: got NULL client in redraw_client_frame!\n",PROGNAME);
#endif
  if(c==NULL) return;

#ifdef DEBUG 
  printf("  redraw_client_frame...  [frame: %#lx] window: %#lx\n",c->frame,c->window); 
#endif 
 
#ifdef SHOW_WEIRD 
  if(!c->has_frame) printf("%s oddity: client/window (%#lx,%#lx) has no frame - no frame drawing done\n",PROGNAME,c,c->window); 
#endif 
  if(!c->has_frame) return; 

  if(c==get_focus_client() && gX11InFront!=0) 
    state=FRAME_ACTIVE;

/*  if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED)*/
  if( c->is_icon==ICONBAR_RAISED ) {
    shaded=2;
    state=FRAME_ACTIVE;
  } else if( c->is_icon==ICONBAR_HIDDEN || c->is_icon==ICONBAR_VISIBLE ) {
    shaded=2;
    state=FRAME_INACTIVE;
  }
 
  if(c->is_icon!=NOT_ICONBAR && TW<1) {
    // this iconbar is 'hidden' (i.e. it is shrunk to a small tab on one side of the screen)
    if(iconbarpos==ICONBARPOS_TOPLEFT || iconbarpos==ICONBARPOS_BOTLEFT) {
      // iconbars on left, so try to only do the right side of the iconbar tab
      if(c->top_right) {
        XSetWindowBackgroundPixmap(dpy, c->top_right, config.top_right.pixmap[state+shaded]);
        XClearWindow(dpy, c->top_right);
      } else if(c->top_left) {
        XSetWindowBackgroundPixmap(dpy, c->top_left, config.top_left.pixmap[state+shaded]);
        XClearWindow(dpy, c->top_left);
      }
    } else {
      // iconbars on right, so try to only do the left side of the iconbar tab
      if(c->top_left) {
        XSetWindowBackgroundPixmap(dpy, c->top_left, config.top_left.pixmap[state+shaded]);
        XClearWindow(dpy, c->top_left);
      } else if(c->top_right) {
        XSetWindowBackgroundPixmap(dpy, c->top_right, config.top_right.pixmap[state+shaded]);
        XClearWindow(dpy, c->top_right);
      }
    }
  } else {
    if(c->top_left) {
      XSetWindowBackgroundPixmap(dpy, c->top_left, config.top_left.pixmap[state+shaded]); 
      XClearWindow(dpy, c->top_left); }
    if(c->top_right) {
      XSetWindowBackgroundPixmap(dpy, c->top_right, config.top_right.pixmap[state+shaded]);
      XClearWindow(dpy, c->top_right); }
  }
  
  if( c->is_icon==NOT_ICONBAR ) {

    if(c->left) { XSetWindowBackgroundPixmap(dpy, c->left, config.left.pixmap[state]); 
    XClearWindow(dpy, c->left); }
    if(c->right) { XSetWindowBackgroundPixmap(dpy, c->right, config.right.pixmap[state]); 
    XClearWindow(dpy, c->right); }
    if(c->bottom_left) { XSetWindowBackgroundPixmap(dpy, c->bottom_left, config.bottom_left.pixmap[state]); 
    XClearWindow(dpy, c->bottom_left); }
    if(c->bottom) { XSetWindowBackgroundPixmap(dpy, c->bottom, config.bottom.pixmap[state]); 
    XClearWindow(dpy, c->bottom); }
    if(c->resize) { XSetWindowBackgroundPixmap(dpy, c->resize, config.resize.pixmap[state]); 
    XClearWindow(dpy, c->resize); }
 
    redraw_client_buttons(c); 
  }

  if( TW>0 ) redraw_client_title(c);
  
  XFlush(dpy);
  XSync(dpy,0);
} 
 
/* 
 *  
 * Redraw client buttons 
 * 
 */ 

void redraw_client_buttons(Client *c) 
{ 
  int state=FRAME_INACTIVE; 
 
  sprintf(lastsub,"TRACE3: redraw_client_buttons (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: redraw_client_buttons (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
#endif

#ifdef SHOW_WEIRD
  if(c==NULL) printf("%s oddity: got NULL client in redraw_client_buttons!\n",PROGNAME);
#endif
  if(c==NULL) return;

#ifdef DEBUG
  printf("redraw_client_buttons [frame: %#lx] window: %#lx\n",c->frame,c->window); 
#endif 

#ifdef SHOW_WEIRD 
  if(c->is_icon!=NOT_ICONBAR) printf("%s oddity: client/window (%#lx,%#lx) is an iconbar - no button drawing done\n",PROGNAME,c,c->window);
  if(!c->has_frame) printf("%s oddity: client/window (%#lx,%#lx) has no frame - no title drawing done\n",PROGNAME,c,c->window); 
#endif 
  if( c->is_icon!=NOT_ICONBAR ) return;
  if(!c->has_frame) return;
 
  if(c==get_focus_client() && gX11InFront!=0) 
    state=FRAME_ACTIVE; 
 
  if(c->icon) { XSetWindowBackgroundPixmap(dpy, c->icon, config.icon.pixmap[c->icon_state==BUTTON_UNPRESSED?state:c->icon_state]); 
  XClearWindow(dpy, c->icon); }
  if(c->close) { XSetWindowBackgroundPixmap(dpy, c->close, config.close.pixmap[c->close_state==BUTTON_UNPRESSED?state:c->close_state]); 
  XClearWindow(dpy, c->close); }
  if(c->can_max) {
    if(c->hide_state!=BUTTON_UNPRESSED) state=c->hide_state;
  } else if(state==FRAME_ACTIVE)
    state=BUTTON_ACTIVEOFF;
  else if(state==FRAME_INACTIVE)
    state=BUTTON_INACTIVEOFF;
  if(c->hide) { XSetWindowBackgroundPixmap(dpy, c->hide, config.hide.pixmap[state]);
  XClearWindow(dpy, c->hide); }

} 
 
/* 
 * 
 * Redraw client name in the title bar 
 * 
 */ 

void redraw_client_title(Client *c) 
{ 
  int state=FRAME_INACTIVE, shaded=0; 
  Pixmap temp; 
  GC gc; 
  XGCValues gv; 
#ifdef I18N_SUPPORT 
  XRectangle ink, logical; 
  int ascent=0, descent=0; 
#endif /* I18N */ 
  int width; 
 
  sprintf(lastsub,"TRACE3: redraw_client_title (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: redraw_client_title (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
#endif

#ifdef SHOW_WEIRD
  if(c==NULL)
    printf("%s oddity: got NULL client in redraw_client_title!\n",PROGNAME);
#endif
  if(c==NULL) return;

#ifdef SHOW_WEIRD
  if(c->title==0) printf("%s oddity: got zero title window in redraw_client_title!\n",PROGNAME);
#endif
  if(c->title==0) return;
  
#ifdef DEBUG 
  printf("redraw_client_title [%#lx] %#lx (%#lx)\n",c->frame,c->window,lastfocus); 
#endif 
 
#ifdef SHOW_WEIRD 
  if(!c->has_frame && (c->is_icon==NOT_ICONBAR)) printf("%s oddity: client/window (%#lx,%#lx) has no frame - no title drawing done\n",PROGNAME,c,c->window); 
#endif 
  if(!c->has_frame && (c->is_icon==NOT_ICONBAR)) return;

  if(TW<1) printf("Not drawn title of client %#lx: %d\n",c,TW);
  if(TW<1) return;
  
  if(c==get_focus_client() && gX11InFront!=0) 
    state=FRAME_ACTIVE; 

/*  if(get_gnome_hint(c->window, WIN_STATE)&WIN_STATE_SHADED)*/
  if( c->is_icon==ICONBAR_VISIBLE || c->is_icon==ICONBAR_HIDDEN ) {
    shaded=2;
    state=FRAME_INACTIVE;
  } else if( c->is_icon==ICONBAR_RAISED ) {
    shaded=2;
    state=FRAME_ACTIVE;
  }
 
  gv.function=GXcopy; 
  gv.tile=config.title.pixmap[state+shaded]; 
  gv.fill_style=FillTiled; 
  gc=XCreateGC(dpy, root, GCFunction|GCFillStyle|GCTile, &gv); 
  if( c->is_icon!=NOT_ICONBAR ) {
    temp=XCreatePixmap(dpy, root, TW, TH, DefaultDepth(dpy, screen));
    XFillRectangle(dpy, temp, gc, 0, 0, TW, TH); 
  } else {
    temp=XCreatePixmap(dpy, root, TW, TH, DefaultDepth(dpy, screen)); 
    XFillRectangle(dpy, temp, gc, 0, 0, TW, TH); 
  }
 
  if(c->name) 
  { 
    int posx=0;
    int posy=0; 
#ifdef I18N_SUPPORT
    if(use_fontset)
    {
      XmbTextExtents(config.fontset, c->name, strlen(c->name), &ink, &logical); 
      ascent=-logical.y; 
      descent=logical.height+logical.y; 
      width=logical.width; 
    }
    else
      width=XTextWidth(config.font, c->name, strlen(c->name));
#else  
    width=XTextWidth(config.font, c->name, strlen(c->name)); 
#endif /* I18N */ 
 
    switch(config.text_align) 
    { 
    case ALIGNMENT_MIDDLE: 
      posx=width>TW-2?2:TW/2-width/2; 
      break; 
    case ALIGNMENT_RIGHT: 
      posx=width>TW-2?2:TW-width; 
      break; 
    case ALIGNMENT_LEFT: 
      posx=13; 
      break; 
    }
    
    // iconbars have text on the left - also, deal with case where there is not enough space
    if( c->is_icon!=NOT_ICONBAR )
       posx=3;
    else if( posx<config.close.width+config.hide.width+config.icon.width-1 )
       posx=config.close.width+config.hide.width+config.icon.width-1;

#ifdef I18N_SUPPORT
    if(use_fontset)
      posy=(TH-descent)-((TH-ascent-descent)/2);
    else
      posy=(TH-config.font->descent)-((TH-config.font->ascent-config.font->descent)/2);
#else 
    posy=(TH-config.font->descent)-((TH-config.font->ascent-config.font->descent)/2); 
#endif /* I18N */ 

    if( c->is_icon!=NOT_ICONBAR ) {
      if( c->icontxt_w==width ) {
        // text width has not changed, so leave position as it was
         posx=c->icontxt_x;
         posy=c->icontxt_y;
      } else {
         c->icontxt_x=posx;
         c->icontxt_y=posy;
         c->icontxt_w=width; }
    } else if( c->name_w==width && c->name_x==posx && c->name_y==posy ) {
        // text width has not changed, so leave position as it was
         posx=c->name_x;
         posy=c->name_y;
    } else {
         c->name_x=posx;
         c->name_y=posy;
         c->name_w=width;
    }

#ifdef I18N_SUPPORT
    if(use_fontset)
      XmbDrawString(dpy, temp, config.fontset, config.font_gc[state], posx, posy, c->name, strlen(c->name));
    else
      XDrawString(dpy, temp, config.font_gc[state], posx, posy, c->name, strlen(c->name));
#else 
    XDrawString(dpy, temp, config.font_gc[state], posx, posy, c->name, strlen(c->name)); 
#endif /* I18N */ 

 // need to make sure dock icon's window list is up to date
    if(c->osxwinfo.winref!=0)
      CheckOSXWinTitle(&(c->osxwinfo),c->name,c->clientclass);
  }
  XSetWindowBackgroundPixmap(dpy, c->title, temp); 
  XClearWindow(dpy, c->title); 
 
  XFreeGC(dpy, gc); 
  XFreePixmap(dpy, temp); 
} 
 
#ifdef SHAPE_SUPPORT 
/* 
 * 
 * Set client window shape 
 * 
 */ 

void set_client_shape(Client *c, int inhibit) 
{ 
  Window temp,wintmp;
  int i;
 
  sprintf(lastsub,"TRACE3: set_client_shape (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: set_client_shape (client,lastfocus: %#lx,%#lx)\n",c,lastfocus);
#endif

#ifdef SHOW_WEIRD
  if(c==NULL) printf("%s oddity: got NULL client in set_client_shape!\n",PROGNAME);
#endif
  if(c==NULL) return;

#ifdef DEBUG
  printf("set_client_shape [frame: %#lx] window: %#lx\n",c->frame,c->window); 
#endif 
 
  if(shape) 
//  if(False) 
  { 
    temp=XCreateSimpleWindow(dpy, root, 0, 0, FW, FH, 0, 0, 0);

    if(!c->has_frame && (c->is_icon==NOT_ICONBAR))
    {
      if(inhibit==0)
        XShapeCombineShape(dpy, temp, ShapeBounding, LW, TH, c->window, ShapeBounding, ShapeSet);
    }
    else if( c->is_icon!=NOT_ICONBAR)
    {
        // Just shape titlebar parts for an iconbar
        if(c->top_left) XShapeCombineMask(dpy, c->top_left, ShapeBounding, 0, 0, config.top_left.mask[1], ShapeSet);
        if(c->top_right) XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[1], ShapeSet);
        if(config.title.width>0 && config.title.height>0 && TRX>TX) {
          XShapeCombineMask(dpy, c->title, ShapeBounding, 0, 0, config.title.mask[1], ShapeSet);
          for(i=config.title.width;i<TRX-TX;i+=config.title.width)
            XShapeCombineMask(dpy, c->title, ShapeBounding, i, 0, config.title.mask[1], ShapeUnion);
          wintmp=XCreateSimpleWindow(dpy, root, 0, 0, TRX-TX, TH, 0, 0, 0);
          XShapeCombineShape(dpy, c->title, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
          XDestroyWindow(dpy, wintmp);
        }
        if(c->top_left) XShapeCombineShape(dpy, temp, ShapeBounding, TLX, TLY, c->top_left, ShapeBounding, ShapeSet);    
        if(c->top_right) XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, ShapeUnion);    
        if(c->title) XShapeCombineShape(dpy, temp, ShapeBounding, TX, TY, c->title, ShapeBounding, ShapeUnion);
    }
    else if( (get_gnome_hint(c->window, WIN_STATE) & WIN_STATE_SHADED)==WIN_STATE_SHADED )
    { 
        // Just shape titlebar parts for a shaded window
        if(c->top_left) XShapeCombineMask(dpy, c->top_left, ShapeBounding, 0, 0, config.top_left.mask[0], ShapeSet);
        if(c->top_right) XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[0], ShapeSet);
        if(config.title.width>0 && config.title.height>0 && TRX>TX) {
          XShapeCombineMask(dpy, c->title, ShapeBounding, 0, 0, config.title.mask[0], ShapeSet);
          for(i=config.title.width;i<TRX-TX;i+=config.title.width)
            XShapeCombineMask(dpy, c->title, ShapeBounding, i, 0, config.title.mask[0], ShapeUnion);
          wintmp=XCreateSimpleWindow(dpy, root, 0, 0, TRX-TX, TH, 0, 0, 0);
          XShapeCombineShape(dpy, c->title, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
          XDestroyWindow(dpy, wintmp);
        }
        if(c->top_left) XShapeCombineShape(dpy, temp, ShapeBounding, TLX, TLY, c->top_left, ShapeBounding, ShapeSet);    
        if(c->top_right) XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, ShapeUnion);    
        if(c->title) XShapeCombineShape(dpy, temp, ShapeBounding, TX, TY, c->title, ShapeBounding, ShapeUnion);
    }
    else if(inhibit==0)
    {
        // Make sure the window itself has the right shape
        XShapeCombineShape(dpy, temp, ShapeBounding, LW, TH, c->window, ShapeBounding, ShapeSet);
        // Shape the window corners in case there are 'internal' shapes (eg, triangular resize widget)
        if(c->bottom_left) XShapeCombineMask(dpy, c->bottom_left, ShapeBounding, 0, 0, config.bottom_left.mask[0], ShapeSet);
        if(c->resize) XShapeCombineMask(dpy, c->resize, ShapeBounding, 0, 0, config.resize.mask[0], ShapeSet);
        if(c->top_left) XShapeCombineMask(dpy, c->top_left, ShapeBounding, 0, 0, config.top_left.mask[0], ShapeSet);
        if(c->top_right) XShapeCombineMask(dpy, c->top_right, ShapeBounding, 0, 0, config.top_right.mask[0], ShapeSet);
        if(c->bottom_left) XShapeCombineMask(dpy, c->bottom_left, ShapeBounding, 0, 0, config.bottom_left.mask[0], ShapeSet);
        // Shape the contents of the window widgets
        XShapeCombineMask(dpy, c->hide, ShapeBounding, 0, 0, config.hide.mask[0], ShapeSet);
        XShapeCombineMask(dpy, c->icon, ShapeBounding, 0, 0, config.icon.mask[0], ShapeSet);
        XShapeCombineMask(dpy, c->close, ShapeBounding, 0, 0, config.close.mask[0], ShapeSet);
        // Only shape the frame if we're using a shaped theme...
        if(gUseShapedTheme!=0) {
          // Shape the left edge (this will suffer from the same problem as the title shaping below
          // but works OK for current themes because the bottom_left shape goes right to the edges)
          if(config.left.width>0 && config.left.height>0 && BLY>LY && c->left) {
            XShapeCombineMask(dpy, c->left, ShapeBounding, 0, 0, config.left.mask[0], ShapeSet);
            for(i=config.left.height;i<BLY-LY;i+=config.left.height)
              XShapeCombineMask(dpy, c->left, ShapeBounding, 0, i, config.left.mask[0], ShapeUnion);
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, LW, BLY-LY, 0, 0, 0);
            XShapeCombineShape(dpy, c->left, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
            XDestroyWindow(dpy, wintmp);
          }
          // Shape the right edge (this will suffer from the same problem as the title shaping below
          // but works OK for current themes because the resize shape goes right to the edges)
          if(config.right.width>0 && config.right.height>0 && RSY>RY && c->right) {
            XShapeCombineMask(dpy, c->right, ShapeBounding, 0, 0, config.right.mask[0], ShapeSet);
            for(i=config.right.height;i<RSY-RY;i+=config.right.height)
              XShapeCombineMask(dpy, c->right, ShapeBounding, 0, i, config.right.mask[0], ShapeUnion);
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, RW, RSY-RY, 0, 0, 0);
            XShapeCombineShape(dpy, c->right, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
            XDestroyWindow(dpy, wintmp);
          }
          // Shape the bottom edge (this will suffer from the same problem as the title shaping below
          // but works OK for current themes because the resize shape goes right to the edges)
          if(config.bottom.width>0 && config.bottom.height>0 && RSX>BX && c->bottom) {
            XShapeCombineMask(dpy, c->bottom, ShapeBounding, 0, 0, config.bottom.mask[0], ShapeSet);
            for(i=config.bottom.width;i<RSX-BX;i+=config.bottom.width)
              XShapeCombineMask(dpy, c->bottom, ShapeBounding, i, 0, config.bottom.mask[0], ShapeUnion);
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, RSX-BX, BH, 0, 0, 0);
            XShapeCombineShape(dpy, c->bottom, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
            XDestroyWindow(dpy, wintmp);
          }
          // Shape the title -doesn't work correctly for current themes since it goes too far and
          // interferes with the top_right rounded part of the frame...
          if(config.title.width>0 && config.title.height>0 && TRX>TX) {
            XShapeCombineMask(dpy, c->title, ShapeBounding, 0, 0, config.title.mask[0], ShapeSet);
            for(i=config.title.width;i<TRX-TX;i+=config.title.width)
              XShapeCombineMask(dpy, c->title, ShapeBounding, i, 0, config.title.mask[0], ShapeUnion);
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, TRX-TX, TH, 0, 0, 0);
            XShapeCombineShape(dpy, c->title, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeIntersect); 
            XDestroyWindow(dpy, wintmp);
          }
        }
        else
        {
          if(config.left.width>0 && config.left.height>0 && BLY>LY && c->left) {
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, LW, BLY-LY, 0, 0, 0);
            XShapeCombineShape(dpy, c->left, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeSet); 
            XDestroyWindow(dpy, wintmp);
          }
          if(config.right.width>0 && config.right.height>0 && RSY>RY && c->right) {
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, RW, RSY-RY, 0, 0, 0);
            XShapeCombineShape(dpy, c->right, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeSet); 
            XDestroyWindow(dpy, wintmp);
          }
          if(config.bottom.width>0 && config.bottom.height>0 && RSX>BX && c->bottom) {
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, RSX-BX, BH, 0, 0, 0);
            XShapeCombineShape(dpy, c->bottom, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeSet); 
            XDestroyWindow(dpy, wintmp);
          }
          if(config.title.width>0 && config.title.height>0 && TRX>TX) {
            wintmp=XCreateSimpleWindow(dpy, root, 0, 0, TRX-TX, TH, 0, 0, 0);
            XShapeCombineShape(dpy, c->title, ShapeBounding, 0, 0, wintmp, ShapeBounding, ShapeSet); 
            XDestroyWindow(dpy, wintmp);
          }
        }
        // Now apply all of those shapes to the actual window frame
        if(c->title) XShapeCombineShape(dpy, temp, ShapeBounding, TX, TY, c->title, ShapeBounding, ShapeUnion);  
        if(c->left) XShapeCombineShape(dpy, temp, ShapeBounding, LX, LY, c->left, ShapeBounding, ShapeUnion); 
        if(c->right) XShapeCombineShape(dpy, temp, ShapeBounding, RX, RY, c->right, ShapeBounding, ShapeUnion); 
        if(c->bottom) XShapeCombineShape(dpy, temp, ShapeBounding, BX, BY, c->bottom, ShapeBounding, ShapeUnion);
        if(c->top_left) XShapeCombineShape(dpy, temp, ShapeBounding, TLX, TLY, c->top_left, ShapeBounding, ShapeUnion);    
        if(c->top_right) XShapeCombineShape(dpy, temp, ShapeBounding, TRX, TRY, c->top_right, ShapeBounding, ShapeUnion);    
        if(c->resize) XShapeCombineShape(dpy, temp, ShapeBounding, RSX, RSY, c->resize, ShapeBounding, ShapeUnion);    
        if(c->bottom_left) XShapeCombineShape(dpy, temp, ShapeBounding, BLX, BLY, c->bottom_left, ShapeBounding, ShapeUnion);
    } 
    XShapeCombineShape(dpy, c->frame, ShapeBounding, 0, 0, temp, ShapeBounding, ShapeSet); 
    XDestroyWindow(dpy, temp);
  } 
} 
#endif /* SHAPE */ 
 
#ifdef WORKSPACE_SUPPORT 
/* 
 * 
 * Toggle sticky/unsticky 
 * 
 */ 

void toggle_client_sticky(Client *c) 
{ 
  sprintf(lastsub,"TRACE2: toggle_client_sticky (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: toggle_client_sticky (client,lastfocus: %#lx,%#lx)\n",c,lastfocus); 
#endif 
 
#ifdef DEBUG 
  printf("toggle_client_sticky [frame: %#lx] window:%#lx (%#lx)\n",c->frame,c->window); 
#endif 
 
  set_gnome_hint(c->window, WIN_STATE, get_gnome_hint(c->window, WIN_STATE)^WIN_STATE_STICKY); 
} 
 
/* 
 * 
 * Move client to new workspace 
 * 
 */ 

void move_client_to_workspace(Client *c, int ws) 
{ 
  sprintf(lastsub,"TRACE2: move_client_to_workspace (client,ws: %#lx,%d)\n",c,ws); 
  subtrace();
#ifdef SHOW_TRACE2 
  printf("TRACE2: move_client_to_workspace (client,ws: %#lx,%d)\n",c,ws); 
#endif 

#ifdef SHOW_WEIRD
  if(ws==workspace || ws<0 || ws>=config.workspace_count)
    printf("%s oddity: got invalid workspace in move_client_to_workspace: %d\n",PROGNAME,ws);
#endif

  if(ws==workspace || ws<0 || ws>=config.workspace_count) return; 
  set_gnome_hint(c->window, WIN_WORKSPACE, ws); 
  change_to_workspace(ws); 
  if(c->is_icon==NOT_ICONBAR) {
    set_focus_client(c); 
    raise_client(c); }
} 
#endif /* WORKSPACE */ 

void manipulate_focus_client(int action)
{
  Client *c, *tc;
  
  sprintf(lastsub,"TRACE1: manipulate_focus_client (action: %d)\n",action); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: manipulate_focus_client (action: %d)\n",action); 
#endif 

  c=get_focus_client();
  if(c==NULL) {
#ifdef SHOW_WEIRD
    printf("%s oddity: Got NULL on Window menu: %d\n",PROGNAME,action);
#endif
    return; }
  
  switch(action)
  {
    case itemWindowshade:
      toggle_client_shaded(c);
      break;
      
    case itemMinimise:
      if(c->is_icon==NOT_ICONBAR)
        minmax_client(c);
#ifdef SHOW_WEIRD
      else
        printf("%s oddity: Tried to minimise an iconbar!\n",PROGNAME);
#endif
      break;
      
    case itemZoom:
      toggle_client_maximized(c);
      break;
      
    case itemClose:
      close_client(c);
      break;
      
    case itemLower:
      tc=get_bottom_most_client(get_gnome_hint(c->window,WIN_LAYER),False);
      if(tc!=c) {
        lower_client(c);
        tc=get_top_most_client(get_gnome_hint(c->window,WIN_LAYER),False);
        if(tc!=c && tc!=NULL)
          set_focus_client(tc);
      }
      break;

    case itemNext:
      get_next_osxwinref(1);
//      printf("Next is not yet functional\n");
//      lower_client(c);
      break;

    case itemAutoFocus+100:
//#ifdef SHOW_WEIRD
//      printf("Autofocus not yet available\n");
//#endif
      if((c->auto_properties & AUTO_FOCUS)==0)
        c->auto_properties += AUTO_FOCUS;
      else
        c->auto_properties -= AUTO_FOCUS;
      set_focus_client(c);
      break;
      
    case itemAutoRaise+100:
//#ifdef SHOW_WEIRD
//      printf("Autoraise not yet available\n");
//#endif
      if((c->auto_properties & AUTO_RAISE)==0)
        c->auto_properties += AUTO_RAISE;
      else
        c->auto_properties -= AUTO_RAISE;
      set_focus_client(c);
      break;
      
    case itemFloat+100:
#ifdef SHOW_WEIRD
      printf("Floating palettes not yet available\n");
#endif
      break;

    case itemPopup+100:
#ifdef SHOW_WEIRD
      printf("Popup windows not yet available\n");
#endif
      break;

    case itemNoframes+100:
#ifdef SHOW_WEIRD
      printf("No frames toggle not yet available\n");
#endif
      break;

    default:
#ifdef SHOW_WEIRD
      printf("%s oddity: Got funny action in manipulate_focus_client: %d\n",PROGNAME,action);
#endif
      break;
  }
}

void change_iconbar_menuspace()
{
  Client *c;
  int neary, newx, newy;
  long i;
  
  sprintf(lastsub,"TRACE1: change_iconbar_menuspace (gMenubarSpace: %d)\n",gMenubarSpace); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: change_iconbar_menuspace (gMenubarSpace: %d)\n",gMenubarSpace); 
#endif 

  if(gMenubarSpace==0)
    gMenubarSpace=21;
  else
    gMenubarSpace=-21;

  if(iconbarpos==ICONBARPOS_BOTLEFT || iconbarpos==ICONBARPOS_BOTRIGHT) {
    if(gMenubarSpace==-21) gMenubarSpace=0;
    return;
  }

  neary=nexticon_y;
  
  init_nearest_iconbar();
  c=get_nearest_iconbar(neary,NOT_ICONBAR);
  while(c!=NULL) {
    newx=c->icon_x;
    newy=c->icon_y+gMenubarSpace;
    if(c->is_icon!=NOT_ICONBAR)
      move_client(c,newx,newy);
    else {
      c->icon_x=newx;
      c->icon_y=newy;
    }
    c=get_nearest_iconbar(neary,NOT_ICONBAR);
  }
    
  nexticon_y+=gMenubarSpace;
  
  if(gMenubarSpace==-21) gMenubarSpace=0;
}

void change_iconbar_screen(int newhead)
{
  Client *c;
  int neary, newx, newy, dx, dy;
  
  sprintf(lastsub,"TRACE1: change_iconbar_screen (newhead,current: %d,%d)\n",newhead,iconbar_screen); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: change_iconbar_screen (newhead,current: %d,%d)\n",newhead,iconbar_screen); 
#endif 

  if(iconbar_screen==newhead) return;
  
  neary=nexticon_y;
    
  dx = screen_x(newhead) - screen_x(iconbar_screen);
  dy = screen_y(newhead) - screen_y(iconbar_screen);
  if(iconbarpos==ICONBARPOS_TOPRIGHT) {
    dx += screen_width(newhead) - screen_width(iconbar_screen);
  } else if(iconbarpos==ICONBARPOS_BOTLEFT) {
    dy += screen_height(newhead) - screen_height(iconbar_screen);
  } else if(iconbarpos==ICONBARPOS_BOTRIGHT) {
    dx += screen_width(newhead) - screen_width(iconbar_screen);
    dy += screen_height(newhead) - screen_height(iconbar_screen);
  }
  
  nexticon_x+=dx;
  nexticon_y+=dy;
  
  iconbar_screen=newhead;
  
  init_nearest_iconbar();
  c=get_nearest_iconbar(neary,NOT_ICONBAR);
  
  while(c!=NULL) {
  
    newx=c->icon_x+dx;
    newy=c->icon_y+dy;
    
    if(c->is_icon!=NOT_ICONBAR)
      move_client(c,newx,newy);
    else {
      c->icon_x=newx;
      c->icon_y=newy;
    }
      
    c=get_nearest_iconbar(neary,NOT_ICONBAR);
  }

}

void change_iconbarpos(int newpos)
{
  Client *c;
  int head=0,swapr2l=0, swapl2r=0, swapt2b=0, swapb2t=0, neary, newx, newy;
  
  sprintf(lastsub,"TRACE1: change_iconbarpos (newpos: %d)\n",newpos); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: change_iconbarpos (newpos: %d)\n",newpos); 
#endif 

  if(iconbarpos==newpos) return;
  
  head=iconbar_screen;

//#ifdef XINERAMA_SUPPORT
//    if(xinerama)
//      head=find_head(50, 50);
//#endif /* XINERAMA */

  if(iconbarpos==ICONBARPOS_TOPRIGHT || iconbarpos==ICONBARPOS_TOPLEFT) {
    if(newpos==ICONBARPOS_BOTRIGHT || newpos==ICONBARPOS_BOTLEFT)
      swapt2b=1;
  } else if(newpos==ICONBARPOS_TOPRIGHT || newpos==ICONBARPOS_TOPLEFT)
    swapb2t=1;
  
  if(iconbarpos==ICONBARPOS_TOPLEFT || iconbarpos==ICONBARPOS_BOTLEFT) {
    if(newpos==ICONBARPOS_TOPRIGHT || newpos==ICONBARPOS_BOTRIGHT)
      swapl2r=1;
  } else if(newpos==ICONBARPOS_TOPLEFT || newpos==ICONBARPOS_BOTLEFT)
    swapr2l=1;

  iconbarpos=newpos;
  
  neary=nexticon_y;
  if(swapt2b==1) {
    nexticon_y=screen_y(head)+screen_height(head)-neary+screen_y(head);
    if(gMenubarSpace==0) nexticon_y-=22;
  } else if(swapb2t==1) {
    nexticon_y=screen_y(head)+screen_height(head)-neary+screen_y(head);
    if(gMenubarSpace==0) nexticon_y-=22;
  }
  
  if(swapl2r==1) {
    nexticon_x=screen_x(head)+screen_width(head)-1;
  } else if(swapr2l==1) {
    nexticon_x=screen_x(head);
  }
  
  init_nearest_iconbar();
  c=get_nearest_iconbar(neary,NOT_ICONBAR);
  
  while(c!=NULL) {
  
    newx=c->icon_x;
    newy=c->icon_y;
    
    if(swapl2r==1)
      newx=screen_x(head)+screen_width(head)-c->icon_w-1;
    else if(swapr2l==1)
      newx=screen_x(head);
    
    if(swapt2b==1) {
      newy=screen_y(head)+screen_height(head)-c->icon_y+screen_y(head);
      if(gMenubarSpace==0) newy-=22;
    } else if(swapb2t==1) {
      newy=screen_y(head)+screen_height(head)-c->icon_y+screen_y(head);
      if(gMenubarSpace==0) newy-=22;
    }
    
    if(c->is_icon!=NOT_ICONBAR) {
      move_client(c,newx,newy);
    } else {
      c->icon_x=newx;
      c->icon_y=newy;
    }
      
    c=get_nearest_iconbar(neary,NOT_ICONBAR);
  }
  
}

void change_options(int newattr)
{
  switch(newattr)
  {
    case itemHideIconbars:
#ifdef SHOW_WEIRD
      printf("Should not get itemHideIconbars!\n");
#endif
      Setup_Options_Menu(0);
      break;
      
    case itemTopLeft:
      change_iconbarpos(ICONBARPOS_TOPLEFT);
      Setup_Options_Menu(newattr);
      break;

    case itemTopRight:
      change_iconbarpos(ICONBARPOS_TOPRIGHT);
      Setup_Options_Menu(newattr);
      break;

    case itemBotLeft:
      change_iconbarpos(ICONBARPOS_BOTLEFT);
      Setup_Options_Menu(newattr);
      break;

    case itemBotRight:
      change_iconbarpos(ICONBARPOS_BOTRIGHT);
      Setup_Options_Menu(newattr);
      break;
    
    case itemSlowIconbars:
      iconbar_steps=ICONSTEPS_SLOW;
      Setup_Options_Menu(newattr);
      break;
      
    case itemMedIconbars:
      iconbar_steps=ICONSTEPS_MED;
      Setup_Options_Menu(newattr);
      break;
      
    case itemFastIconbars:
      iconbar_steps=ICONSTEPS_FAST;
      Setup_Options_Menu(newattr);
      break;
      
    case itemFancyIconbars:
      if(shrinking_iconbars==0)
        shrinking_iconbars=1;
      else
        shrinking_iconbars=0;
      Setup_Options_Menu(0);
      break;
      
    case itemInterleave:
#ifdef SHOW_WEIRD
      printf("Automatic interleave control not yet available...\n");
#endif
      Setup_Options_Menu(0);
      break;

    case itemLeaveMenubar:
      change_iconbar_menuspace();
      Setup_Options_Menu(0);
      break;
    
    default:
      if(newattr>=itemScreenListStart) {
        change_iconbar_screen(newattr-itemScreenListStart);
        Setup_Options_Menu(newattr);
      } else {
#ifdef SHOW_WEIRD
      printf("%s oddity: Got unknown attr in change_options\n",PROGNAME);
#endif
      }
      break;
  }

}

void unmap_all_clients() {
  Window *windows=NULL, w1, w2; 
  int count, i; 
  Client *c;
  
  focus_before_hide=get_focus_client();
  
  sprintf(lastsub,"TRACE1: unmap_all_clients\n"); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE1: unmap_all_clients\n"); 
#endif
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
#ifdef DEBUG
  printf("  children of root: %d\n",count);
#endif

  for(i=0;i<count;i++) 
  { 
#ifdef DEBUG
    printf("  child %d (%#lx)\n",i,windows[i]);
#endif
    c=get_client_from_window(windows[i],FRAME);
    if(c!=NULL) {
      c->ignore_unmap++; 
      XUnmapWindow(dpy, c->frame); 
      XUnmapWindow(dpy, c->window); 
      set_wm_state(c->window, WithdrawnState); 
    }
  } 

  if(windows) XFree(windows);
  
//  c=client_list;
//  for(i=0;i<client_count;i++) {
//    c=c->next;
//  }
}

/*
 * For debugging purposes
 */
void list_clients_and_windows()
{
  Window *windows=NULL; 
  int count, i; 
  Client *c;

  c=client_list;
  if(c==NULL) return;
  if(client_count==0) {
    printf("%s oddity: No clients in list, but client_list is %#lx !?\n",PROGNAME,client_list);
    return;
  }
  
  printf("%s info: List of clients (name, client, frame, window, title, top_left, top_right, left, right, bottom, bottom_left, resize, close, icon, max):\n",PROGNAME);
  for(i=0;i<client_count;i++) {
    printf(" ""%s"" %#lx, %#lx, %#lx, %#lx, %#lx, %#lx, %#lx, %#lx, %#lx, %#lx, %#lx, %#lx, %#lx, %#lx\n",c->name,c,c->frame,c->window,c->title,c->top_left,c->top_right,c->left,c->right,c->bottom,c->bottom_left,c->resize,c->close,c->icon,c->hide);
    c=c->next;
  }
}


void map_all_clients() {
  Window *windows=NULL, w1, w2; 
  int count, i; 
  Client *c;
    
  sprintf(lastsub,"TRACE1: map_all_clients\n"); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE1: map_all_clients\n"); 
#endif
 
  XQueryTree(dpy, root, &w1, &w2, &windows, &count); 
#ifdef DEBUG
  printf("  children of root: %d\n",count);
#endif

  for(i=0;i<count;i++) 
  { 
#ifdef DEBUG
    printf("  child %d (%#lx)\n",i,windows[i]);
#endif
    c=get_client_from_window(windows[i],FRAME);
    if(c!=NULL) {
      c->ignore_unmap++; 
      XMapWindow(dpy, c->frame); 
      XMapWindow(dpy, c->window); 
      set_wm_state(c->window, NormalState); 
    }
  } 

  if(windows) XFree(windows);

//  c=client_list;
//  for(i=0;i<client_count;i++) {
//    c=c->next;
//  }

  if(focus_before_hide!=NULL)
    set_focus_client(focus_before_hide);
  focus_before_hide=NULL;
}

/*
 * Check that OSX windows are off-screen
 */
void check_osx_windows_bounds() {
  Window *windows=NULL; 
  int count, i; 
  Client *c;

    sprintf(lastsub,"TRACE3: check_osx_windows_bounds\n");
    subtrace();
#ifdef SHOW_TRACE3
    printf("TRACE3: check_osx_windows_bounds\n");
#endif

  c=client_list;
  if(c==NULL) return;
  if(client_count==0) {
    printf("%s oddity: No clients in list, but client_list is %#lx !?\n",PROGNAME,client_list);
    return;
  }
  
  for(i=0;i<client_count;i++) {
    if(c->osxwinfo.winref!=0)
        Check_OSX_Window_Bounds(&c->osxwinfo);
    c=c->next;
  }
}

/*
 * Find the OSX WindowRef (returned as a long) for the bottom-most client (for the "Next" menu item)
 */
long get_next_osxwinref(int raiseit)
{
    Client *c, *fc;
    long osxwr=0;
    
    fc=get_focus_client();
    
#ifdef DEBUG
    if(fc==NULL)
      printf("Got NULL focus client in get_next_osxwinref.\n");
#endif
    if(fc==NULL) return 0;
    
    c=get_bottom_most_client(get_gnome_hint(fc->window,WIN_LAYER),False);
    
#ifdef DEBUG
    if(c==fc)
      printf("Got focus client is next client in get_next_osxwinref.\n");
#endif
    if(c==fc) return 0;
    
    if(c!=NULL) {
      if(raiseit!=0) {
        set_focus_client(c);
        raise_client(c);
      }
      osxwr=c->osxwinfo.winref;
    }
    
    sprintf(lastsub,"TRACE2: get_next_osxwinref -got next client: %#lx (winref: %d)\n",c,osxwr);
    subtrace();
#ifdef SHOW_TRACE2
    printf("TRACE3: get_next_osxwinref -got next client: %#lx (winref: %d)\n",c,osxwr);
#endif
  
    return osxwr;
}


