/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Sept 2001      */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 

/* common.h */

/*
 * This contains the info that both the 'oroborus part' and thhe 'Carbon part'
 * of OroborOSX need to know about
 */
 
#include <unistd.h>
#include "definitions.h"

#define PROGNAME		"OroborOSX"

#define MOUSE_STOPPED	-1
#define MOUSE_NORMAL	0
#define MOUSE_MOVE	1
#define MOUSE_RESIZE	2
#define MOUSE_CLOSE	3
#define MOUSE_HIDE	4
#define MOUSE_CLOSE_OUT	5
#define MOUSE_HIDE_OUT	6
#define MOUSE_OVERCLOSE	7
#define MOUSE_OVERHIDE	8
#define MOUSE_ICON	9
#define MOUSE_ICON_OUT	10

#define ICONBARPOS_TOPLEFT	0
#define ICONBARPOS_TOPRIGHT	1
#define ICONBARPOS_BOTLEFT	2
#define ICONBARPOS_BOTRIGHT	3

#define	menuApplication		128	/* Application menu ID */
#define	itemAbout		1

#define menuLaunch		129	/* Launch menu ID */
#define itemEditStart		1
#define itemNewItem		3
#define itemEditItems		4
#define itemRebuildMenu		5
#define itemxinitrc		7
// points to divider item in Launch menu before the items added from the Launch Menu Items folder
#define itemUserItems		8
// These have now been replaced by x11app items in the Launch menu Items folder...
/*
#define itemxterm		9
#define itemxedit		10
#define itemxbiff		11
#define itemxclock		12
#define itemxeyes		13
#define itemxcalc		14
#define itemUserItems		15
// The rest of these are not actually used - they are just to remind of the default for the user part
#define itemETerm		16
#define itemGimp		17
#define itemNEdit		18
*/

#define	menuEdit		130	/* Edit menu ID */

#define	menuOptions		131	/* Options menu ID */
#define itemInterleave		1
#define itemThemeSubmenu	3
#define itemLeaveMenubar	3
#define itemHideIconbars	5
#define itemFancyIconbars	6
#define itemSlowIconbars	8
#define itemMedIconbars		9
#define itemFastIconbars	10
#define itemTopLeft		12
#define itemTopRight		13
#define itemBotLeft		14
#define itemBotRight		15
#define itemScreenListStart	17

#define menuThemes		132	/* Themes submenu ID */
#define itemAboutThemes		1
#define itemAddTheme		3
#define itemEditThemeList	4
#define itemRebuildThemes	5
#define itemToggleShaped	7
#define itemAboutThisTheme	9
// points to the divider in the Themes submenu before the items added from the Themes folder
#define itemUserThemes		10
// These items are not actually used -they are replaced by the themes that are actually found...
#define itemEau			11
#define itemGreyphite		12
#define itemNextish		13

#define	menuWindow		133	/* Window menu ID */
#define itemWindowshade		1
#define	itemMinimise		2
#define	itemZoom		3
#define	itemClose		5
#define itemProperties		7
#define itemLower		9
#define itemNext		10

#define menuProperties		135	/* Properties submenu ID */
#define itemAutoFocus		1
#define itemAutoRaise		2
#define itemFloat		4
#define itemPopup		5
#define itemNoframes		7

#define	menuHelp		134	/* Help menu ID */
#define itemHelp		1

// Extra window info structure
typedef struct 
{
    long winref;
    int x, y, width, height;
    int widgets_x, widgets_y, widgets_w, widgets_h;
} OSXWinInfo; 

/* globals */
extern int realPathLen;
extern char realPath[512];
extern int on_cmd_line, shrinking_iconbars, iconbar_screen, live_resizing;
extern int gQuitFlag, gX11InFront, gInterleaving_Windows;
extern int gHiding_Iconbars, gMenubarSpace, gUseShapedTheme, gX11MouseStuck;
extern int mouse_mode, real_x, real_y, iconbarpos, diffscrn_x, diffscrn_y;
extern long lastBackTicks, lastIconbarOutTicks;
extern char *Display_Name, *xpidstr, *xverstr, *gThemeDir, lastsub[255];
extern char *lastsub15, *lastsub14, *lastsub13, *lastsub12, *lastsub11;
extern char *lastsub10, *lastsub9, *lastsub8, *lastsub7, *lastsub6;
extern char *lastsub5, *lastsub4, *lastsub3, *lastsub2, *lastsub1;

/* main.c */
extern void Bring_To_Front(int quick);
extern void Bring_X_To_Front(int quick);
extern void Bring_OOSX_To_Front();
extern void MakeX11Window(OSXWinInfo *winfo, char *title, char *class);
extern void FocusOSXWin(OSXWinInfo *winfo);
extern void DestroyOSXWindow(OSXWinInfo *winfo);
extern void CheckOSXWinTitle(OSXWinInfo *winfo, char *title, char *class);
extern void HideOSXWin(OSXWinInfo *winfo);
extern void ReorderOSXWin(OSXWinInfo *winfo);
extern void SetOSXWinPos(OSXWinInfo *winfo,int x, int y, int w, int h, int wox, int woy, int wendox, int wendoy);
extern void will_hide_all_iconbars();
extern void Setup_Options_Menu(int which);
extern void Setup_Window_Menu(int has_frame, int has_close, int has_min, int has_zoom, int is_float, int can_af, int can_ar);
extern void Compare_Screens(int scrnx, int scrny, int x11x, int x11y, int allowset);
extern void Stop_XDarwin();
extern void Restart_XDarwin();
extern void subtrace();
extern void Check_OSX_Window_Bounds(OSXWinInfo *winfo);
extern void show_last_fifteen();
extern void show_last_ten();

/* oroborus.c */
extern void quit();
extern int orob_main();

/* events.c */ 
extern void do_event_loop(int sleepus,int trackMouse);
extern void force_ungrab();

/* client.c */
extern void left_client();
extern void gone_to_back();
extern void come_to_front();
extern void change_focus(long winref);
extern void hide_all_iconbars();
extern void manipulate_focus_client(int action);
extern void change_options(int newattr);
extern void raise_all_clients();
extern void map_all_clients();
extern void unmap_all_clients();
extern void check_osx_windows_bounds();
extern void list_clients_and_windows(); 
extern long get_next_osxwinref(int);

/* config.c */ 
extern void reload_config(); 

/* misc.c */
extern int count_xinerama_heads();

