/*************************************/ 
/* Original oroborus window manager  */
/* (c) Ken Lynch Jan 2001            */
/*                                   */
/* Extra functionality for OroborOSX */
/* (c) Adrian Umpleby Sept 2001      */
/*                                   */
/* Distributed under the GPL         */ 
/* See LICENSE for more details      */ 
/*************************************/ 
 
#include "oroborus.h" 
 
/* 
 * 
 * Print out X errors 
 * 
 */ 
int handle_xerror(Display *dpy, XErrorEvent *err) 
{ 
  char msg[1024]; 
 
  sprintf(lastsub,"TRACE1: handle_xerror (code: %d)\n",err->error_code); 
  subtrace();
#ifdef SHOW_TRACE1 
  printf("TRACE1: handle_xerror (code: %d)\n",err->error_code); 
#endif 
 
  switch(err->error_code) 
  { 
  case BadAccess:  
    if(err->resourceid==root)
    {
      if(on_cmd_line) {
        fprintf(stderr, "%s: Another window manger is running.\n", PROGNAME);
        exit(1);
      } else {
        gQuitFlag=-4;
        return 0;
      }
    }
    break; 
  case BadWindow: 
    XGetErrorText(dpy, err->error_code, msg, 1024);
#ifdef SHOW_WEIRD
    if(!inhibit_errors) {
      fprintf(stderr, "%s: X BadWindow Warning (%#lx): %s.\n", PROGNAME, err->resourceid, msg);
      show_last_ten(); }
#endif
    break;
  case BadDrawable: 
    XGetErrorText(dpy, err->error_code, msg, 1024);
#ifdef SHOW_WEIRD
    fprintf(stderr, "%s: X BadDrawable Warning (%#lx): %s.\n", PROGNAME, err->resourceid, msg);
    show_last_ten();
#endif
    break;
  case BadMatch: 
    XGetErrorText(dpy, err->error_code, msg, 1024);
#ifdef SHOW_WEIRD
    fprintf(stderr, "%s: X BadMatch Warning (%#lx): %s.\n", PROGNAME, err->resourceid, msg);
    show_last_ten();
#endif
    break; 
  default: 
    XGetErrorText(dpy, err->error_code, msg, 1024);
//    if(on_cmd_line)
      fprintf(stderr, "%s: X Error (number %#lx): %s.\n", PROGNAME, err->resourceid, msg);
//    else
//      return 0;
    show_last_ten();
  } 
  return 0; 
} 

/*
 *
 * Add directory to font path
 *
 */

void add_dir_to_font_path(char *font_dir)
{
  char **font_path, **new_font_path, *file_name;
  int npaths;
  FILE *file;

  sprintf(lastsub,"TRACE1: add_dir_to_font_path\n");
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: add_dir_to_font_path\n");
#endif

  if((file_name=malloc(strlen(font_dir)+strlen("fonts.dir")+2))==NULL)
  {
    if(on_cmd_line) {
      fprintf(stderr, "%s: Memory allocation failed in function add_dir_to_font_path.\n", PROGNAME);
      exit(1);
    } else {
      fprintf(stderr, "%s: Memory allocation failed in function add_dir_to_font_path.\n", PROGNAME);
      gQuitFlag=-5;
      return;
    }
  }

  if((file=fopen(file_name, "r")))
  {
    fclose(file);
    font_path=XGetFontPath(dpy, &npaths);
    if((new_font_path=malloc(sizeof(char *)*(npaths+1)))==NULL)
    {
      if(on_cmd_line) {
        fprintf(stderr, "%s: Memory allocation failed in function add_dir_to_font_path.\n", PROGNAME);
        exit(1);
      } else {
        fprintf(stderr, "%s: Memory allocation failed in function add_dir_to_font_path.\n", PROGNAME);
        gQuitFlag=-6;
        return;
      }
    }
    memcpy(new_font_path, font_path, sizeof(char*)*npaths);
    new_font_path[npaths]=font_dir;
    XSetFontPath(dpy, new_font_path, npaths+1);
    XFreeFontPath(font_path);
    free(new_font_path);
  }
}

/*
 *
 * Remove directory from font path
 *
 */

void rem_dir_from_font_path(char *font_dir)
{
  char **font_path, **new_font_path;
  int npaths, i, j=0;

  sprintf(lastsub,"TRACE1: rem_dir_from_font_path\n");
  subtrace();
#ifdef SHOW_TRACE1
  printf("TRACE1: rem_dir_from_font_path\n");
#endif

  font_path=XGetFontPath(dpy, &npaths);
  if((new_font_path=malloc(sizeof(char *)*npaths))==NULL)
  {
    if(on_cmd_line) {
      fprintf(stderr, "%s: Memory allocation failed in function rem_dir_from_font_path.\n", PROGNAME);
      exit(1);
    } else {
      fprintf(stderr, "%s: Memory allocation failed in function rem_dir_from_font_path.\n", PROGNAME);
      gQuitFlag=-7;
      return;
    }
  }
  for(i=0;i<npaths;i++)
  {
    if(strcmp(font_path[i], font_dir))
      new_font_path[j++]=font_path[i];
  }
  XSetFontPath(dpy, new_font_path, j);
  XFreeFontPath(font_path);
  free(new_font_path);
}

/* 
 * 
 * Get mouse x any y 
 * 
 */ 

void get_mouse_xy(int *x, int *y) 
{ 
  Window w1, w2; 
  int wx, wy, mask; 
 
  sprintf(lastsub,"TRACE3: get_mouse_xy\n"); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_mouse_xy\n"); 
#endif 
  if((gX11InFront==0 && mouse_mode==MOUSE_NORMAL) || replay_if_drag) {
    // if XDarwin is behind, then it does not track the cursor - so use values set by main
    *x=real_x;
    *y=real_y;
//    XWarpPointer(dpy, None, root, 0, 0, 0, 0, *x, *y);
  } else {
    XQueryPointer(dpy, root, &w1, &w2, x, y, &wx, &wy, &mask);
//    if(mouse_mode==MOUSE_NORMAL && ((w2!=root && w2!=0) || client_list==NULL))
    if(mouse_mode==MOUSE_NORMAL) {
      Compare_Screens(real_x,real_y,*x,*y,1);	// want to check to see how X11 and 'real' screen compare
      if(*x!=real_x || *y!=real_y) {
        printf("   changing cursor pos in get_mouse: %d,%d  to  %d,%d",*x,*y,real_x,real_y);
        *x=real_x;
        *y=real_y;
      }
    } else if(*x!=real_x || *y!=real_y) {
        printf("   cannot change cursor pos in get_mouse (%d): %d,%d  cf  %d,%d",mouse_mode,*x,*y,real_x,real_y);
    }
  }
} 
 
/* 
 * 
 * Get window under pointer 
 * 
 */ 

Window get_window_container(int x, int y, Window w)
{
  Window *windows=NULL, w1, w2, top=0; 
  int count, i; 
  XWindowAttributes attr; 
 
  sprintf(lastsub,"TRACE3: get_window_container (w: %#lx  x,y: %d,%d)\n",w,x,y); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: get_window_container (w: %#lx  x,y: %d,%d)\n",w,x,y); 
#endif
 
  XGetWindowAttributes(dpy, w, &attr); 
  x=x-attr.x;
  y=y-attr.y;

  sprintf(lastsub,"  doing XQueryTree in get_window_container\n"); 
  subtrace();
  XQueryTree(dpy, w, &w1, &w2, &windows, &count); 
#ifdef DEBUG
  printf("  attr x,y,w,h & children: %d,%d %d,%d  %d\n",attr.x,attr.y,attr.width,attr.height,count);
#endif

  sprintf(lastsub,"  done XQueryTree in get_window_container: %d (%#lx)\n",count,windows); 
  subtrace();
  
  for(i=count-1;i>=0;i--) 
  { 
    XGetWindowAttributes(dpy, windows[i], &attr);
#ifdef DEBUG
    printf("  child %d (%#lx): %d,%d %d,%d\n",i,windows[i],attr.x,attr.y,attr.width,attr.height);
#endif
    if(attr.map_state==IsViewable)
      if(x>=attr.x)
        if(x<attr.x+attr.width)
          if(y>=attr.y)
            if(y<attr.y+attr.height) {
              top=windows[i];
              break; }
  } 

  if(windows) XFree(windows);
  
  sprintf(lastsub,"TRACE3: leaving get_window_container (top: %#lx)\n",top); 
  subtrace();
#ifdef SHOW_TRACE3
  printf("TRACE3: leaving get_window_container (top: %#lx)\n",top); 
#endif
 
  return top; 
}

Window get_pointer_win(Window w) 
{ 
  Window w1=0, w2=0, w3=0; 
  int x, y, wx, wy, mask; 
 
  sprintf(lastsub,"TRACE3: get_pointer_win (w: %#lx)\n",w); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_pointer_win (w: %#lx)\n",w); 
#endif 

  if((gX11InFront!=0 && gX11MouseStuck==0) || mouse_mode!=MOUSE_NORMAL || replay_if_drag) {
    if(w==root) {
      w3=get_window_container(real_x,real_y,w);
      if(w3==0) {
//        printf("++ cursor in window: %#lx (%#lx)\n",w3,w3,lastenter);
        if(mouse_mode==MOUSE_NORMAL) {
          XQueryPointer(dpy, w, &w1, &w2, &x, &y, &wx, &wy, &mask);
          // want to check to see how X11 and 'real' screen compare
          if(w2==0)
            Compare_Screens(real_x,real_y,x,y,1);
          else
            Compare_Screens(real_x,real_y,x,y,0);
        }
  sprintf(lastsub,"TRACE3: leaving get_pointer_win with nothing\n"); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: leaving get_pointer_win with nothing\n"); 
#endif 
        return 0;
      }
    }
    XQueryPointer(dpy, w, &w1, &w2, &x, &y, &wx, &wy, &mask);
//    if(w==root && mouse_mode==MOUSE_NORMAL && ((w2!=root && w2!=0) || client_list==NULL))
//    if(w==root) {
//      printf("++ cursor in window: %#lx, %#lx (%#lx)\n",w2,w3,lastenter);
//    }
    if(w==root && mouse_mode==MOUSE_NORMAL) {
      // want to check to see how X11 and 'real' screen compare
      if(w2==w3)
         Compare_Screens(real_x,real_y,x,y,1);
      else
         Compare_Screens(real_x,real_y,x,y,0);
    }
  } else {    // if XDarwin is behind, then it does not track the cursor - so use x,y values set by main
    w2=get_window_container(real_x,real_y,w);
//    if(w==root) {
//      printf("-- cursor in window: %#lx (%#lx, %#lx)\n",w2,lastenter,check_win);
//    }
  }
  
  sprintf(lastsub,"TRACE3: leaving get_pointer_win (w2: %#lx)\n",w2); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: leaving get_pointer_win (w2: %#lx)\n",w2); 
#endif 

  return w2; 
} 
 
/* 
 * 
 * Get window parent 
 * 
 */ 

Window get_window_parent(Window win) 
{ 
  Window parent, dummyw, *windows=NULL; 
  int count; 
 
  sprintf(lastsub,"TRACE3: get_window_parent (win: %#lx)\n",win); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: get_window_parent (win: %#lx)\n",win); 
#endif 
 
  XQueryTree(dpy, win, &dummyw, &parent, &windows, &count); 
  if(windows) XFree(windows); 
  return parent; 
} 
 
/* 
 * 
 * Screen width 
 * 
 */ 
int screen_width(int head) 
{ 
  sprintf(lastsub,"TRACE3: screen_width (head: %d)\n",head); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: screen_width (head: %d)\n",head); 
#endif 
 
#ifdef XINERAMA_SUPPORT 
  if(xinerama) 
    return xinerama_screens[head].width; 
#endif /* XINERAMA */ 
  return XDisplayWidth(dpy, screen); 
} 
 
/* 
 * 
 * Screen height 
 * 
 */ 

int screen_height(int head) 
{ 
  sprintf(lastsub,"TRACE3: screen_height (head: %d)\n",head); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: screen_height (head: %d)\n",head); 
#endif 
 
#ifdef XINERAMA_SUPPORT 
  if(xinerama) 
    return xinerama_screens[head].height; 
#endif /* XINERAMA */ 
  return XDisplayHeight(dpy, screen); 
} 
 
/* 
 * 
 * Screen x 
 * 
 */ 

int screen_x(int head) 
{ 
  sprintf(lastsub,"TRACE3: screen_x (head: %d)\n",head); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: screen_x (head: %d)\n",head); 
#endif 
 
#ifdef XINERAMA_SUPPORT 
  if(xinerama) 
    return xinerama_screens[head].x_org; 
#endif /* XINERAMA */ 
  return 0; 
} 
 
/* 
 * 
 * Screen y 
 * 
 */ 

int screen_y(int head) 
{ 
  sprintf(lastsub,"TRACE3: screen_y (head: %d)\n",head); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: screen_y (head: %d)\n",head); 
#endif 
 
#ifdef XINERAMA_SUPPORT 
  if(xinerama) 
    return xinerama_screens[head].y_org; 
#endif /* XINERAMA */ 
  return 0; 
} 
 
#ifdef XINERAMA_SUPPORT 
/*
 * glue routine for main to get count of xinerama heads
 */

int count_xinerama_heads()
{
  return xinerama_heads;
}
 
/* 
 * 
 * Find head 
 * 
 */ 

int find_head(int x, int y) 
{ 
  int head, cx, cy, dx, dy, min_dist_sqr, dist_sqr, closest_head; 
 
  sprintf(lastsub,"TRACE3: find_head (x,y: %d,%d)\n",x,y); 
  subtrace();
#ifdef SHOW_TRACE3 
  printf("TRACE3: find_head (x,y: %d,%d)\n",x,y); 
#endif 
 
  for(head=0;head<xinerama_heads;head++) 
  { 
    if(x>=xinerama_screens[head].x_org && x<xinerama_screens[head].x_org+xinerama_screens[head].width && y>=xinerama_screens[head].y_org && y<xinerama_screens[head].y_org+xinerama_screens[head].height) 
      return head; 
  } 
  
#ifdef DEBUG 
  printf("  no head found, looking for nearest head...\n"); 
#endif 
  
  /* No head found, return closest */ 
  cx=xinerama_screens[0].x_org+(xinerama_screens[0].width/2); 
  cy=xinerama_screens[0].y_org+(xinerama_screens[0].height/2); 
 
  dx=x-cx; 
  dy=y-cy; 
  min_dist_sqr=(dx*dx)+(dy*dy); 
  closest_head=0; 
 
  for(head=1;head<xinerama_heads;head++) 
  { 
    cx=xinerama_screens[head].x_org+(xinerama_screens[head].width/2); 
    cy=xinerama_screens[head].y_org+(xinerama_screens[head].height/2); 
 
    dx=x-cx; 
    dy=y-cy; 
    dist_sqr=(dx*dx)+(dy*dy); 
     
    if(dist_sqr<min_dist_sqr) 
    { 
      closest_head=head; 
      min_dist_sqr=dist_sqr; 
    } 
  } 
 
  return closest_head; 
} 
#endif /* XINERAMA */ 
 
#ifdef I18N_SUPPORT 
/* 
 * 
 * some wrapper functions for I18N 
 * 
 */ 

Status XGetWMNameWrapper(Display *dpy, Window window, XTextProperty *text_prop) 
{  
  int status; 
  char **list; 
  int num; 
  
  status = XGetWMName(dpy, window, text_prop); 
  if (!status || !text_prop->value || text_prop->nitems <= 0) 
    return 0; 
  if (text_prop->encoding == XA_STRING) 
    return 1; 
   
  text_prop->nitems = strlen((char *)text_prop->value); 
  status = XmbTextPropertyToTextList(dpy, text_prop, &list, &num); 
   
  if (status >= Success && num > 0 && *list)
  { 
    if(text_prop->value) XFree(text_prop->value); 
    text_prop->value = (unsigned char*)strdup(*list); 
    text_prop->nitems = strlen(*list); 
    if(list) XFreeStringList(list); 
    return 1; 
  } 
   
  return 0; 
} 
 
Status XFetchNameWrapper(Display *dpy, Window window, char **name) 
{  
  XTextProperty text_prop; 
   
  if (XGetWMNameWrapper(dpy, window, &text_prop) != 0)
  { 
    if (text_prop.value && text_prop.nitems > 0)
    { 
      *name = (char *)text_prop.value; 
      return 1; 
    } 
  } 
  *name = NULL; 
  return 0; 
} 
 
XFontSet XLoadQueryFontWrapper(Display *dpy, const char *font_name) 
{ 
  char **miss, *def=NULL; 
  int n_miss; 
  char *base_font_name_list=NULL; 
  int font_name_list_length; 
  XFontSet _fontset; 
 
  font_name_list_length = strlen(font_name) + 3; 
  base_font_name_list = (char *) malloc(sizeof(char) * font_name_list_length); 
  if (!base_font_name_list)
  { 
    if(on_cmd_line) {
      fprintf(stderr, "%s: Memory allocation failed in function XLoadQueryFontWrapper.\n", PROGNAME);
      exit(1);
    } else {
      fprintf(stderr, "%s: Memory allocation failed in function XLoadQueryFontWrapper.\n", PROGNAME);
      gQuitFlag=-8;
      return NULL;
    }
  }
  sprintf(base_font_name_list, "%s,*", font_name); 
  _fontset = XCreateFontSet(dpy, base_font_name_list, &miss, &n_miss, &def); 
  if (!_fontset)
  {
    free(base_font_name_list);
    return NULL;
  }
  if (n_miss)
  { 
    int j; 
    if(on_cmd_line) {
      fprintf(stderr, "%s: XLoadQueryFontWrapper(): missing charset(s)\n", PROGNAME); 
      for (j=0; j<n_miss; j++)  
        fprintf(stderr, "\t%s\n", miss[j]); 
    }
    XFreeStringList(miss); 
  } 
  free(base_font_name_list);
  return _fontset; 
} 
#endif /* I18N */ 
